import numpy as np
from GridCal.Engine.basic_structures import Logger, SolverType, ReactivePowerControlMode, TapsControlMode
from GridCal.Engine.Core.multi_circuit import MultiCircuit
from GridCal.Engine.basic_structures import BranchImpedanceMode
from GridCal.Engine.basic_structures import BusMode
from GridCal.Engine.Devices.enumerations import ConverterControlType, TransformerControlType
from GridCal.Engine.Simulations.PowerFlow.power_flow_options import PowerFlowOptions

try:
    import NewtonNative as nn
    NEWTON_AVAILBALE = True
    print('Newton Native v' + nn.get_version())

except ImportError:
    NEWTON_AVAILBALE = False
    newton_solver_dict = dict()
    newton_taps_dict = dict()
    newton_q_control_dict = dict()
    print('Newton native is not available')


def set_branch_values(nc: "nn.NativeNumericCircuit",
                      idx, name, active, f, t, rate, r, x, g, b, m, theta, vtap_f, vtap_t):
    """
    Set the the main parameters of a branch
    :param nc: NativeNumericCircuit
    :param idx: branch index
    :param name: name of the branch
    :param active: is active?
    :param f: "from" bus index
    :param t: "to" bus index
    :param rate: branch rating
    :param r: resistance (p.u.)
    :param x: reactance (p.u.)
    :param g: conductance (p.u.)
    :param b: susceptance (p.u.)
    :param m: tap module
    :param theta: tap angle
    :param vtap_f: virtual tap "from"
    :param vtap_t: virtual tap "to"
    """
    nc.set_branch_name(idx, name)
    nc.set_branch_active(idx, active)
    nc.set_branch_rates(idx, rate)

    nc.set_branch_r(idx, r)
    nc.set_branch_x(idx, x)
    nc.set_branch_g(idx, g)
    nc.set_branch_b(idx, b)
    nc.set_branch_tap_mod(idx, m)
    nc.set_branch_tap_ang(idx, theta)
    nc.set_branch_tap_f(idx, vtap_f)
    nc.set_branch_tap_t(idx, vtap_t)

    nc.set_c_branch_bus_f(idx, f, 1)
    nc.set_c_branch_bus_t(idx, t, 1)
    nc.set_branch_f(idx, f)
    nc.set_branch_t(idx, t)


def to_newton_native(circuit: MultiCircuit) -> "nn.NativeNumericCircuit":
    """
    Compile the information of a circuit and generate the pertinent power flow islands
    :param circuit: MultiCircuit instance
    :return: NativeNumericCircuit
    """

    bus_dictionary = dict()

    # Element count
    nbus = len(circuit.buses)
    nload = 0
    ngen = 0
    n_batt = 0
    nshunt = 0
    for bus in circuit.buses:
        nload += len(bus.loads)
        ngen += len(bus.controlled_generators)
        n_batt += len(bus.batteries)
        nshunt += len(bus.shunts)

    nline = len(circuit.lines)
    ndcline = len(circuit.dc_lines)
    ntr2w = len(circuit.transformers2w)
    # ntr3w = len(circuit.transformers3w)
    nvsc = len(circuit.vsc_devices)
    nhvdc = len(circuit.hvdc_lines)

    # declare the numerical circuit
    nc = nn.NativeNumericCircuit(nbus=nbus,
                                 nline=nline,
                                 ndcline=ndcline,
                                 ntr=ntr2w,  # + 3 * ntr3w,
                                 nvsc=nvsc,
                                 nhvdc=nhvdc,
                                 nload=nload,
                                 ngen=ngen,
                                 nbatt=n_batt,
                                 nshunt=nshunt,
                                 sBase=circuit.Sbase)

    # buses and it's connected elements (loads, generators, etc...)
    i_ld = 0
    i_gen = 0
    i_batt = 0
    i_sh = 0
    V0 = np.ones(nbus, dtype=complex)
    for i, bus in enumerate(circuit.buses):

        # bus parameters
        nc.set_bus_name(i, bus.name)
        nc.set_bus_active(i, bus.active)
        nc.set_bus_types(i, bus.determine_bus_type().value)

        # Add buses dictionary entry
        bus_dictionary[bus.idtag] = i

        # initialize voltage guess
        V0[i] = bus.get_voltage_guess()

        for elm in bus.loads:
            nc.set_load_name(i_ld, elm.name)
            nc.set_load_active(i_ld, elm.active)
            nc.set_load_s(i_ld, complex(elm.P, elm.Q))
            nc.set_c_bus_load(i, i_ld, 1)
            i_ld += 1

        for elm in bus.controlled_generators:
            nc.set_generator_name(i_gen, elm.name)
            nc.set_generator_active(i_gen, elm.active)
            nc.set_generator_controllable(i_gen, elm.is_controlled)
            nc.set_generator_pf(i_gen, elm.Pf)
            nc.set_generator_v(i_gen, elm.Vset)
            nc.set_generator_qmin(i_gen, elm.Qmin)
            nc.set_generator_qmax(i_gen, elm.Qmax)

            nc.set_generator_p(i_gen, elm.P)
            nc.set_generator_installed_p(i_gen, elm.Snom)

            nc.set_c_bus_gen(i, i_gen, 1)

            i_gen += 1

        for elm in bus.batteries:
            nc.set_battery_names(i_batt, elm.name)
            nc.set_battery_p(i_batt, elm.P)
            nc.set_battery_pf(i_batt, elm.Pf)
            nc.set_battery_v(i_batt, elm.Vset)
            nc.set_battery_qmin(i_batt, elm.Qmin)
            nc.set_battery_qmax(i_batt, elm.Qmax)
            nc.set_battery_active(i_batt, elm.active)
            nc.set_battery_controllable(i_batt, elm.is_controlled)
            nc.set_battery_installed_p(i_batt, elm.Snom)

            nc.setC_bus_batt(i, i_batt, 1)

            i_batt += 1

        for elm in bus.shunts:
            nc.set_shunt_name(i_sh, elm.name)
            nc.set_shunt_active(i_sh, elm.active)
            nc.set_shunt_admittance(i_sh, complex(elm.G, elm.B))

            nc.set_c_bus_shunt(i, i_sh, 1)
            i_sh += 1

    for i in range(nbus):
        nc.set_bus_v0(i, complex(V0[i].real, V0[i].imag))

    # Compile the lines
    for i, elm in enumerate(circuit.lines):

        f = bus_dictionary[elm.bus_from.idtag]
        t = bus_dictionary[elm.bus_to.idtag]

        # generic stuff
        set_branch_values(nc=nc, idx=i, name=elm.name,
                          active=elm.active, f=f, t=t, rate=elm.rate,
                          r=elm.R, x=elm.X, g=0, b=elm.B, m=1.0, theta=0.0,
                          vtap_f=1.0, vtap_t=1.0)

        # impedance
        nc.set_line_name(i, elm.name)
        nc.set_line_active(i, elm.active)
        nc.set_line_r(i, elm.R)
        nc.set_line_x(i, elm.X)
        nc.set_line_b(i, elm.B)
        nc.set_line_impedance_tolerance(i, elm.tolerance)
        nc.set_c_line_bus(i, f, 1)
        nc.set_c_line_bus(i, t, 1)

        # Thermal correction
        nc.set_line_temp_base(i, elm.temp_base)
        nc.set_line_temp_oper(i, elm.temp_oper)
        nc.set_line_alpha(i, elm.alpha)

    # 2-winding transformers
    for i, elm in enumerate(circuit.transformers2w):
        ii = i + nline
        tr_tap_f, tr_tap_t = elm.get_virtual_taps()

        # generic stuff
        f = bus_dictionary[elm.bus_from.idtag]
        t = bus_dictionary[elm.bus_to.idtag]

        set_branch_values(nc=nc, idx=ii, name=elm.name,
                          active=elm.active, f=f, t=t, rate=elm.rate,
                          r=elm.R, x=elm.X, g=elm.G, b=elm.B,
                          m=elm.tap_module, theta=elm.angle,
                          vtap_f=tr_tap_f, vtap_t=tr_tap_t)

        # impedance
        nc.set_tr_r(i, elm.R)
        nc.set_tr_x(i, elm.X)
        nc.set_tr_g(i, elm.G)
        nc.set_tr_b(i, elm.B)

        nc.set_c_tr_bus(i, f, 1)
        nc.set_c_tr_bus(i, t, 1)

        # tap changer
        nc.set_tr_tap_mod(i, elm.tap_module)
        nc.set_tr_tap_ang(i, elm.angle)
        nc.set_tr_is_bus_to_regulated(i, elm.bus_to_regulated)
        nc.set_tr_tap_position(i, elm.tap_changer.tap)
        nc.set_tr_min_tap(i, elm.tap_changer.min_tap)
        nc.set_tr_max_tap(i, elm.tap_changer.max_tap)
        nc.set_tr_tap_inc_reg_up(i, elm.tap_changer.inc_reg_up)
        nc.set_tr_tap_inc_reg_down(i, elm.tap_changer.inc_reg_down)
        nc.set_tr_vset(i, elm.vset)

        # virtual taps for transformers where the connection voltage is off
        nc.set_tr_tap_f(i, tr_tap_f)
        nc.set_tr_tap_t(i, tr_tap_t)

    # VSC
    for i, elm in enumerate(circuit.vsc_devices):
        ii = i + nline + ntr2w  # + 3 * ntr3w

        # generic stuff
        f = bus_dictionary[elm.bus_from.idtag]
        t = bus_dictionary[elm.bus_to.idtag]

        set_branch_values(nc=nc, idx=ii, name=elm.name,
                          active=elm.active, f=f, t=t, rate=elm.rate,
                          r=elm.R1, x=elm.X1, g=elm.G0, b=elm.Beq,
                          m=elm.m, theta=elm.theta,
                          vtap_f=1.0, vtap_t=1.0)

        # vsc values
        nc.set_vsc_R1(i, elm.R1)
        nc.set_vsc_X1(i, elm.X1)
        nc.set_vsc_Gsw(i, elm.G0)
        nc.set_vsc_Beq(i, elm.Beq)
        nc.set_vsc_m(i, elm.m)
        nc.set_vsc_theta(i, elm.theta)

        nc.set_c_vsc_Bus(i, f, 1)
        nc.set_c_vsc_Bus(i, t, 1)

    # HVDC: this is the simple detached HVDC model
    for i, elm in enumerate(circuit.hvdc_lines):
        ii = i + nline + ntr2w + nvsc

        # generic stuff
        f = bus_dictionary[elm.bus_from.idtag]
        t = bus_dictionary[elm.bus_to.idtag]

        # hvdc values
        nc.set_hvdc_name(i, elm.name)
        nc.set_hvdc_active(i, elm.active)
        nc.set_hvdc_rate(i, elm.rate)

        nc.set_hvdc_pset_f(i, -elm.Pset)
        nc.set_hvdc_pset_t(i, elm.Pset)
        nc.set_hvdc_vset_f(i, elm.Vset_f)
        nc.set_hvdc_vset_t(i, elm.Vset_t)
        nc.set_hvdc_qmin_f(i, elm.Qmin_f)
        nc.set_hvdc_qmax_f(i, elm.Qmax_f)
        nc.set_hvdc_qmin_t(i, elm.Qmin_t)
        nc.set_hvdc_qmax_t(i, elm.Qmax_t)

        # hack the bus types to believe they are PV
        if elm.active:
            nc.set_bus_types(f, BusMode.PV.value)
            nc.set_bus_types(t, BusMode.PV.value)

        # the the bus-hvdc line connectivity
        nc.set_c_hvdc_bus_f(i, f, 1)
        nc.set_c_hvdc_bus_t(i, t, 1)

    return nc


def get_newton_power_flow_options(opt: PowerFlowOptions):
    """

    :param opt:
    :return:
    """
    newton_solver_dict = {SolverType.NR: nn.NativeSolverType.NR,
                          SolverType.DC: nn.NativeSolverType.DC,
                          SolverType.HELM: nn.NativeSolverType.HELM,
                          SolverType.IWAMOTO: nn.NativeSolverType.IWAMOTO,
                          SolverType.LM: nn.NativeSolverType.LM,
                          SolverType.LACPF: nn.NativeSolverType.LACPF,
                          SolverType.FASTDECOUPLED: nn.NativeSolverType.FD}

    newton_taps_dict = {TapsControlMode.NoControl: nn.NativeTapsControlMode.NoControl,
                        TapsControlMode.Direct: nn.NativeTapsControlMode.Direct}

    newton_q_control_dict = {ReactivePowerControlMode.NoControl: nn.NativeReactivePowerControlMode.NoControl,
                             ReactivePowerControlMode.Direct: nn.NativeReactivePowerControlMode.Direct}

    if opt.solver_type in newton_solver_dict.keys():
        solver_type = newton_solver_dict[opt.solver_type]
    else:
        solver_type = nn.NativeSolverType.NR

    options = nn.NativePowerFlowOptions(solver_type=solver_type,
                                        retry_with_other_methods=opt.retry_with_other_methods,
                                        verbose=opt.verbose,
                                        initialize_with_existing_solution=opt.initialize_with_existing_solution,
                                        tolerance=opt.tolerance,
                                        max_iter=opt.max_iter,
                                        control_q_mode=newton_q_control_dict[opt.control_Q],
                                        tap_control_mode=nn.NativeTapsControlMode.NoControl,
                                        distributed_slack=opt.distributed_slack,
                                        ignore_single_node_islands=opt.ignore_single_node_islands,
                                        correction_parameter=opt.backtracking_parameter,
                                        mu0=opt.mu)

    return options


def newton_power_flow(nc: "nn.NativeNumericCircuit", options: PowerFlowOptions):
    """

    :param nc: NativeNumericCircuit instance
    :param options:
    :return:
    """

    options = get_newton_power_flow_options(options)

    # declare the native power flow driver
    native_driver = nn.NativePowerFlow(nc, options)

    # run power flow
    return native_driver.run()


# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
# 
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""
This file includes extensions to the PuLP library
"""

import numpy as np
from .pulp import LpProblem, LpVariable
from itertools import product
from scipy.sparse import csc_matrix


def lpDot(mat, arr):
    """
    CSC matrix-vector or CSC matrix-matrix dot product (A x b)
    :param mat: CSC sparse matrix (A)
    :param arr: dense vector or matrix of object type (b)
    :return: vector or matrix result of the product
    """
    n_rows, n_cols = mat.shape

    # check dimensional compatibility
    assert (n_cols == arr.shape[0])

    # check that the sparse matrix is indeed of CSC format
    if mat.format == 'csc':
        mat_2 = mat
    else:
        # convert the matrix to CSC sparse
        mat_2 = csc_matrix(mat)

    if len(arr.shape) == 1:
        """
        Uni-dimensional sparse matrix - vector product
        """
        res = np.zeros(n_rows, dtype=arr.dtype)
        for i in range(n_cols):
            for ii in range(mat_2.indptr[i], mat_2.indptr[i + 1]):
                j = mat_2.indices[ii]  # row index
                res[j] += mat_2.data[ii] * arr[i]  # C.data[ii] is equivalent to C[i, j]
    else:
        """
        Multi-dimensional sparse matrix - matrix product
        """
        cols_vec = arr.shape[1]
        res = np.zeros((n_rows, cols_vec), dtype=arr.dtype)

        for k in range(cols_vec):  # for each column of the matrix "vec", do the matrix vector product
            for i in range(n_cols):
                for ii in range(mat_2.indptr[i], mat_2.indptr[i + 1]):
                    j = mat_2.indices[ii]  # row index
                    res[j, k] += mat_2.data[ii] * arr[i, k]  # C.data[ii] is equivalent to C[i, j]
    return res


def lpAddRestrictions(problem: LpProblem, arr, name):
    """
    Add vector or matrix of restrictions to the problem
    :param problem: instance of LpProblem
    :param arr: 1D or 2D array
    :param name: name of the restriction
    """

    if len(arr.shape) == 1:

        for i, elm in enumerate(arr):
            problem.add(elm, name + '_' + str(i + 1))

    elif len(arr.shape) == 2:

        for i, j in product(range(arr.shape[0]), range(arr.shape[1])):
            problem.add(arr[i, j], name + '_' + str(i + 1) + '_' + str(j))


def lpAddRestrictions2(problem: LpProblem, lhs, rhs, name, op='='):
    """
    Add vector or matrix of restrictions to the problem
    :param problem: instance of LpProblem
    :param lhs: 1D array (left hand side)
    :param rhs: 1D or 2D array (right hand side)
    :param name: name of the restriction    
    :param op: type of restriction (=, <=, >=)
    """

    assert(lhs.shape == rhs.shape)

    arr = np.empty(lhs.shape, dtype=object)

    if len(lhs.shape) == 1:

        for i in range(lhs.shape[0]):
            if op == '=':
                arr[i] = lhs[i] == rhs[i]

            elif op == '<=':
                arr[i] = lhs[i] <= rhs[i]

            elif op == '>=':
                arr[i] = lhs[i] >= rhs[i]

    elif len(lhs.shape) == 2:

        for i, j in product(range(lhs.shape[0]), range(lhs.shape[1])):
            if op == '=':
                arr[i, j] = lhs[i, j] == rhs[i, j]

            elif op == '<=':
                arr[i, j] = lhs[i, j] <= rhs[i, j]

            elif op == '>=':
                arr[i, j] = lhs[i, j] >= rhs[i, j]

    lpAddRestrictions(problem=problem, arr=arr, name=name)

    return arr


def lpAddRestrictions3(problem: LpProblem, lhs, var, rhs, name):
    """
    Add vector or matrix of restrictions to the problem
    so that lhs <= var <= rhs
    :param problem: instance of LpProblem
    :param lhs: 1D array (left hand side)
    :param var: 1D or 2D array (right hand side)
    :param rhs: 1D or 2D array (right hand side)
    :param name: name of the restriction
    """

    assert (lhs.shape == rhs.shape)
    assert (var.shape == rhs.shape)

    arr_l = np.empty(lhs.shape, dtype=object)
    arr_r = np.empty(lhs.shape, dtype=object)

    if len(lhs.shape) == 1:

        for i in range(lhs.shape[0]):
            arr_r[i] = var[i] <= rhs[i]
            arr_l[i] = lhs[i] <= var[i]

    elif len(lhs.shape) == 2:

        for i, j in product(range(lhs.shape[0]), range(lhs.shape[1])):
            arr_r[i, j] = var[i, j] <= rhs[i, j]
            arr_l[i, j] = lhs[i, j] <= var[i, j]

    lpAddRestrictions(problem=problem, arr=arr_l, name=name + '_l')
    lpAddRestrictions(problem=problem, arr=arr_r, name=name + '_r')

    return arr_l, arr_r


def lpMakeVars(name, shape, lower=None, upper=None):
    """
    Declares 1D of 2D array of LpVars
    :param name: name of the variable
    :param shape: tuple with the shape (i.e. (3), or (4, 6))
    :param lower: Lower bound array. Must meet the shape
    :param upper: Upper bound array. must meet the shape
    :return: array of LpVars
    """

    var = np.empty(shape, dtype=object)

    if type(shape) == int:

        if lower is None and upper is not None:

            for i in range(shape):
                var[i] = LpVariable(name + '_' + str(i), lowBound=None, upBound=upper[i])

        elif lower is not None and upper is None:

            if type(lower) == int:
                for i in range(shape):
                    var[i] = LpVariable(name + '_' + str(i), lowBound=lower, upBound=None)
            else:
                for i in range(shape):
                    var[i] = LpVariable(name + '_' + str(i), lowBound=lower[i], upBound=None)

        elif lower is None and upper is None:
            for i in range(shape):
                var[i] = LpVariable(name + '_' + str(i), lowBound=None, upBound=None)

        else:

            if type(lower) in (float, int) and type(upper) in (float, int):
                for i in range(shape):
                    var[i] = LpVariable(name + '_' + str(i), lowBound=lower, upBound=upper)

            elif type(lower) not in (float, int) and type(upper) in (float, int):
                for i in range(shape):
                    var[i] = LpVariable(name + '_' + str(i), lowBound=lower[i], upBound=upper)

            elif type(lower) in (float, int) and type(upper) not in (float, int):
                for i in range(shape):
                    var[i] = LpVariable(name + '_' + str(i), lowBound=lower, upBound=upper[i])

            elif type(lower) not in (float, int) and type(upper) not in (float, int):
                for i in range(shape):
                    var[i] = LpVariable(name + '_' + str(i), lowBound=lower[i], upBound=upper[i])


            else:
                raise Exception('Cannot handle the indices...')

    else:

        if len(shape) == 1:
            if lower is None and upper is not None:

                for i in range(shape[0]):
                    var[i] = LpVariable(name + '_' + str(i), lowBound=None, upBound=upper[i])

            elif lower is not None and upper is None:

                for i in range(shape[0]):
                    var[i] = LpVariable(name + '_' + str(i), lowBound=lower[i], upBound=None)

            else:

                for i in range(shape[0]):
                    var[i] = LpVariable(name + '_' + str(i), lowBound=lower[i], upBound=upper[i])

        elif len(shape) == 2:

            if lower is None and upper is not None:

                if type(upper) in [float, int]:
                    for i, j in product(range(shape[0]), range(shape[1])):
                        var[i, j] = LpVariable(name + '_' + str(i) + '_' + str(j), lowBound=None, upBound=upper)
                else:
                    for i, j in product(range(shape[0]), range(shape[1])):
                        var[i, j] = LpVariable(name + '_' + str(i) + '_' + str(j), lowBound=None, upBound=upper[i])

            elif lower is not None and upper is None:

                if type(lower) in [float, int]:
                    for i, j in product(range(shape[0]), range(shape[1])):
                        var[i, j] = LpVariable(name + '_' + str(i) + '_' + str(j), lowBound=lower, upBound=None)
                else:
                    for i, j in product(range(shape[0]), range(shape[1])):
                        var[i, j] = LpVariable(name + '_' + str(i) + '_' + str(j), lowBound=lower[i], upBound=None)

            else:

                if type(lower) in [float, int] and type(upper) in [float, int]:
                    for i, j in product(range(shape[0]), range(shape[1])):
                        var[i, j] = LpVariable(name + '_' + str(i) + '_' + str(j), lowBound=lower, upBound=upper)

                elif len(lower.shape) == 2 and len(lower.shape) == 2:
                    for i, j in product(range(shape[0]), range(shape[1])):
                        var[i, j] = LpVariable(name + '_' + str(i) + '_' + str(j), lowBound=lower[i, j], upBound=upper[i, j])

                elif len(lower.shape) == 1 and len(lower.shape) == 2:
                    for i, j in product(range(shape[0]), range(shape[1])):
                        var[i, j] = LpVariable(name + '_' + str(i) + '_' + str(j), lowBound=lower[i], upBound=upper[i, j])

                elif len(lower.shape) == 2 and len(lower.shape) == 1:
                    for i, j in product(range(shape[0]), range(shape[1])):
                        var[i, j] = LpVariable(name + '_' + str(i) + '_' + str(j), lowBound=lower[i, j], upBound=upper[i])

                elif len(lower.shape) == 1 and len(lower.shape) == 1:
                    for i, j in product(range(shape[0]), range(shape[1])):
                        var[i, j] = LpVariable(name + '_' + str(i) + '_' + str(j), lowBound=lower[i], upBound=upper[i])
        else:
            pass

    return var


def lpGet2D(arr, make_abs=False):
    """
    Extract values fro the 2D array of LP variables
    :param arr: 2D array of LP variables
    :param make_abs: substitute the result by its abs value
    :return: 2D numpy array
    """
    val = np.zeros(arr.shape)
    for i, j in product(range(val.shape[0]), range(val.shape[1])):
        val[i, j] = arr[i, j].value()
    if make_abs:
        val = np.abs(val)

    return val



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import os
from datetime import datetime, timedelta
import pandas as pd
import numpy as np
from enum import Enum
import zipfile
from xml.etree import cElementTree as ElementTree
from GridCal.Engine import MultiCircuit, Bus, Generator, Branch, Load, GeneratorTechnologyType, BranchType, FileSave


class XmlDictConfig(dict):
    """
    Note: need to add a root into if no exising
    Example usage:
    >>> tree = ElementTree.parse('your_file.xml')
    >>> root = tree.getroot()
    >>> xmldict = XmlDictConfig(root)
    Or, if you want to use an XML string:
    >>> root = ElementTree.XML(xml_string)
    >>> xmldict = XmlDictConfig(root)
    And then use xmldict for what it is... a dict.
    """
    def __init__(self, parent_element, text_to_remove=''):
        """

        :param parent_element:
        :param text_to_remove:
        """
        self.text_to_remove = text_to_remove

        if parent_element.items():
            self.update_shim(dict(parent_element.items()))

        for element in parent_element:

            tag = element.tag.replace(self.text_to_remove, '')

            if len(element):

                a_dict = XmlDictConfig(element, self.text_to_remove)

                self.update_shim({tag: a_dict})

            elif element.items():    # items() is especialy for attributes

                element_attrib = element.items()

                if element.text:
                    element_attrib.append((tag, element.text))     # add tag:text if there exist

                self.update_shim({tag: dict(element_attrib)})
            else:
                self.update_shim({tag: element.text})

    def update_shim(self, a_dict):
        """

        :param a_dict:
        :return:
        """
        for og_key in a_dict.keys():   # keys() includes tag and attributes

            key = og_key.replace(self.text_to_remove, '')

            if key in self:
                value = self.pop(key)
                if type(value) is not list:
                    list_of_dicts = list()
                    list_of_dicts.append(value)
                    list_of_dicts.append(a_dict[key])
                    self.update({key: list_of_dicts})
                else:
                    value.append(a_dict[key])
                    self.update({key: value})
            else:
                self.update({key: a_dict[key]})  # it was self.update(aDict)


class PlexosBusMode(Enum):
    PQ = 1
    PV = 2
    REF = 3
    NONE = 4
    STO_DISPATCH = 5  # Storage dispatch, in practice it is the same as REF


class PlexosElement:

    def __init__(self, name):
        """
        Generic element constructor
        :param name: Name of the element
        """
        self.name = name

    def __str__(self):
        return self.name


class PlexosNode(PlexosElement):

    def __init__(self, name='', zone='', region='', voltage=0, latitude=0, longitude=0):

        PlexosElement.__init__(self, name=name)

        self.zone = zone

        self.region = region

        self.voltage = voltage

        self.latitude = latitude

        self.longitude = longitude

        self.load = 0.0

        self.load_prof = None

        self.generators = list()

        self.batteries = list()

        self.key = 0  # usually the PSS/e key...

        # get the PSS/e name if that'd be there (PSSNODE_NAME_KV)
        if '_' in self.name:
            self.key = self.name.split('_')[0]
            try:
                self.key = int(self.key)  # try to set it as an integer value
            except ValueError:
                pass


class PlexosGenerator(PlexosElement):

    def __init__(self, name='', category='', p_max=0, p_min=0, node=None):
        PlexosElement.__init__(self, name=name)

        self.category = category

        self.p_max = p_max

        self.p_min = p_min

        self.node = None

        self.rating_factor_prof = None

        self.aux_fixed_prof = None

        self.must_run_units_prof = None


class PlexosBattery(PlexosGenerator):

    def __init__(self, name='', category=''):
        """

        :param name:
        :param category:
        """
        PlexosGenerator.__init__(self, name=name, category=category)


class PlexosLine(PlexosElement):

    def __init__(self, name='', units=1, node_from=None, node_to=None, r=0.0, x=0.0, rate_max=0.0, rate_min=0.0):
        """

        :param name:
        :param units: number of active units (if 0, then the line is not active)
        :param node_from:
        :param node_to:
        :param r:
        :param x:
        :param rate_max:
        :param rate_min:
        """
        PlexosElement.__init__(self, name=name)

        self.units = units

        self.node_from = node_from

        self.node_to = node_to

        self.r = r

        self.x = x

        self.rate_max = rate_max

        self.rate_min = rate_min

        self.coordinates = list()

        self.rate_max_prof = None

        self.rate_min_prof = None

    def get_key(self, sep='-'):
        """
        Split the name in the plexos way to get the new key
        """
        vals = self.name.split('_')
        return vals[0] + sep + vals[3] + sep + vals[6]

    def get_highest_voltage(self):
        """
        Return the highest voltage at which this line is connected
        :return:
        """
        return max(self.node_from.voltage, self.node_to.voltage)

    def delete_zero_coordinates(self):
        """

        :return:
        """
        # delete coordinates with zeros
        for k in range(len(self.coordinates) - 1, -1, -1):
            a, b = self.coordinates[k]
            if a == 0.0 or b == 0.0:
                self.coordinates.pop(k)

    def get_coordinates(self):
        """
        Get polyline of coordinates
        :return:
        """
        if len(self.coordinates) >= 2:
            locations = self.coordinates
        else:
            locations = [[self.node_from.latitude, self.node_from.longitude],
                         [self.node_to.latitude, self.node_to.longitude]]

        # delete coordinates with zeros
        for k in range(len(locations) - 1, -1, -1):
            a, b = locations[k]
            if (a + b) <= 0:
                locations.pop(k)

        return locations


class PlexosTransformer(PlexosLine):

    def __init__(self, name=''):
        PlexosLine.__init__(self, name=name)


class PlexosZone(PlexosElement):

    def __init__(self, name=''):
        PlexosElement.__init__(self, name=name)


class PlexosRegion(PlexosElement):

    def __init__(self, name=''):
        PlexosElement.__init__(self, name=name)


class PlexosModel:

    def __init__(self, fname, load_profiles=True, text_func=None, prog_func=None):
        """
        Plexos model
        :param fname: plexos file name (either .zip or .xml)
        :param load_profiles: load profiles associated
        :param text_func: function pointer to print text
        :param prog_func: function pointer to show the progress
        """

        # name of the project file
        self.file_name = fname
        self.load_profiles = load_profiles
        self.text_func = text_func
        self.prog_func = prog_func

        # directory of this project
        self.directory = os.path.dirname(self.file_name)

        # name of this instance
        self.name = ''

        # dictionary of nodes
        self.nodes = dict()

        # dictionary of generators
        self.generators = dict()

        # dictionary of batteries
        self.batteries = dict()

        # dictionary of lines
        self.lines = dict()

        # dictionary of transformers
        self.transformers = dict()

        # dictionary of lines and transformers all together
        self.branches = dict()

        # dictionary of zones
        self.zones = dict()

        # dictionary of regions
        self.regions = dict()

        # dictionary of transformers od/or lines. dict['line']['line name']
        self.branches_by_type = dict()

        # dictionary of data profiles. The profiles are loaded on the fly
        self.data_profiles = dict()

        # load the project file
        self.load_project_file(fname=self.file_name)

    def load_project_file(self, fname):
        """
        Load a PLEXOS project file
        :param fname: name of the plexos project file (*.zip, *.xml, *.xlsx)
        """
        # split the file name path into name and extension
        filename, file_extension = os.path.splitext(fname)

        # parse the file information
        if file_extension == '.xlsx':
            objects2, memberships2, data = self.parse_excel(fname=self.file_name)
            zip_file_pointer = None

        elif file_extension == '.xml':
            objects2, memberships2, data = self.parse_xml(fname=self.file_name)
            zip_file_pointer = None

        elif file_extension == '.zip':
            objects2, memberships2, data, zip_file_pointer = self.parse_zip(fname=self.file_name)

        else:
            raise Exception('File type not supported: ' + fname)

        # convert the xml data into objects for this class
        self.parse_data(objects=objects2,
                        memberships=memberships2,
                        properties=data,
                        zip_file_pointer=zip_file_pointer)

        # close the zip file
        if zip_file_pointer is not None:
            zip_file_pointer.close()

    def load_profile(self, path, zip_file_pointer=None):
        """
        Attempt loading the profile
        :param path: relative or absolute path
        :param zip_file_pointer: pointer to open zip file is the file ins inside a zip file
        :return: DataFrame
        """
        if zip_file_pointer is not None:
            # create a buffer to read the file
            path2 = path.replace('\\', '/')
            final_path = zip_file_pointer.open(path2)
            df = pd.read_csv(final_path)
            return df

        else:
            if ':' in path:
                final_path = path
            else:
                final_path = os.path.join(self.directory, path)

            if os.path.exists(final_path):
                df = pd.read_csv(final_path)
                return df
            else:
                print(final_path, 'not found :(')
                return None

    def load_profile_if_necessary(self, key, path, zip_file_pointer=None):
        """
        Load a profile is necessary
        :param key: object property type
        :param path: relative or absolute path where to find the file
        :param zip_file_pointer: pointer to open zip file is the file ins inside a zip file
        :return: Nothing
        """

        if key not in self.data_profiles.keys():

            df = self.load_profile(path=path, zip_file_pointer=zip_file_pointer)

            if df is not None:
                self.data_profiles[key] = df
                return df
            else:
                # the profile does not exist
                return False
        else:
            # the profile exists
            return self.data_profiles[key]

    def parse_zip(self, fname):
        """
        Parse zip file with the plexos xml and the profiles utilized
        :param fname: zip file name
        :return: Nothing
        """
        # open the zip file
        zip_file_pointer = zipfile.ZipFile(fname)

        # search for the xml file
        names = zip_file_pointer.namelist()
        xml_file_name = None
        for name in names:
            if name.endswith('.xml'):
                xml_file_name = name
                break

        if xml_file_name is not None:
            # parse the xml and read the data profiles
            objects2, memberships2, data = self.parse_xml(xml_file_name, zip_file_pointer=zip_file_pointer)

        else:
            raise Exception('No xml file was found in the zip file', fname)

        return objects2, memberships2, data, zip_file_pointer

    @staticmethod
    def parse_excel(fname):
        """
        Parse excel export of the plexos file
        :param fname: complete path to the file
        """
        excel = pd.ExcelFile(fname)

        print('Reading objects...')
        objects = excel.parse(sheet_name='Objects')

        print('Reading Memberships...')
        memberships = excel.parse(sheet_name='Memberships')

        print('Reading Properties...')
        properties = excel.parse(sheet_name='Properties')
        properties.rename(columns={'filename': 'path'})
        excel.close()

        # file_dict = {row['child_object']: row['filename'] for i, row in properties.iterrows()}

        return objects, memberships, properties

    def parse_xml(self, fname, zip_file_pointer=None):
        """
        Parse PLEXOS file
        :param fname: xml PLEXOS file name
        :param zip_file_pointer: pointer to a zip file, if not none, the file will be read from within a zip file
        """
        if self.text_func is None:
            print('Parsing plexos xml', fname, '...')
        else:
            self.text_func('Parsing plexos xml ' + fname)

        # read xml tree from file or from zip-file pointer
        if zip_file_pointer is not None:
            file_pointer = zip_file_pointer.open(fname)
            xtree = ElementTree.parse(file_pointer)
        else:
            xtree = ElementTree.parse(fname)

        # get xml root node
        root = xtree.getroot()

        # text to remove: {http://tempuri.org/MasterDataSet.xsd}
        # this is a very annoying text that is present in all the xml nodes
        text_to_remove = root.tag[root.tag.find("{"):root.tag.find("}") + 1]

        # pass the XML file to a dictionary
        xmldict = XmlDictConfig(root, text_to_remove=text_to_remove)

        # pass the dictionaries to Pandas DataFrames
        classes = pd.DataFrame(xmldict['t_class'])
        category = pd.DataFrame(xmldict['t_category'])
        collections = pd.DataFrame(xmldict['t_collection'])

        '''
        objects 
        class	GUID	name	category
        '''
        objects = pd.DataFrame(xmldict['t_object'])
        objects2 = objects.copy()
        objects2['class_id'] = objects2['class_id'].map(classes.set_index('class_id')['name'])
        objects2['category_id'] = objects2['category_id'].map(category.set_index('category_id')['name'])
        objects2.rename(columns={'class_id': 'class', 'category_id': 'category'}, inplace=True)

        '''
        Membership
        parent_class	child_class	   collection	parent_object	child_object
        '''
        memberships = pd.DataFrame(xmldict['t_membership'])

        memberships2 = memberships.copy()
        memberships2['parent_class_id'] = memberships2['parent_class_id'].map(classes.set_index('class_id')['name'])
        memberships2['child_class_id'] = memberships2['child_class_id'].map(classes.set_index('class_id')['name'])
        memberships2['parent_object_id'] = memberships2['parent_object_id'].map(objects.set_index('object_id')['name'])
        memberships2['child_object_id'] = memberships2['child_object_id'].map(objects.set_index('object_id')['name'])
        memberships2['collection_id'] = memberships2['collection_id'].map(
            collections.set_index('collection_id')['name'])
        memberships2.rename(columns={'parent_class_id': 'parent_class',
                                     'child_class_id': 'child_class',
                                     'parent_object_id': 'parent_object',
                                     'child_object_id': 'child_object',
                                     'collection_id': 'collection'}, inplace=True)

        '''
        Properties
        parent_class	child_class	collection	parent_object	child_object	property	band_id	value	units	
        date_from	date_to	pattern	action	expression	filename	scenario	memo	period_type_id
        '''
        attribute = pd.DataFrame(xmldict['t_attribute'])
        attribute_data = pd.DataFrame(xmldict['t_attribute_data'])
        # attributes = pd.merge(attribute, attribute_data, on='attribute_id')

        data = pd.DataFrame(xmldict['t_data'])
        tag = pd.DataFrame(xmldict['t_tag'])
        text = pd.DataFrame(xmldict['t_text'])
        text.rename(columns={'value': 'path'}, inplace=True)

        properties = pd.DataFrame(xmldict['t_property'])

        data = pd.merge(data, memberships2, on='membership_id')
        data = pd.merge(data, text, on='data_id', how='left')
        data = pd.merge(data, tag, on='data_id', how='left')
        data['property_id'] = data['property_id'].map(properties.set_index('property_id')['name'])
        data['object_id'] = data['object_id'].map(objects.set_index('object_id')['name'])
        data.rename(columns={'property_id': 'property', 'object_id': 'filename'}, inplace=True)

        return objects2, memberships2, data

    def parse_data(self, objects, memberships, properties, zip_file_pointer=None):
        """
        Pass the loaded DataFrames to model objects
        :param objects: Objects DataFrame
        :param memberships: Memberships DataFrame
        :param properties: Properties DataFrame
        :param zip_file_pointer: zip file pointer. If not None the data will be read from a zip file
        :return: Nothing
        """
        # create the objects
        if self.text_func is None:
            print('Reading plexos xml file...', end='')
        else:
            self.text_func('Reading plexos xml file... ')

        for i, row in objects.iterrows():

            if row['class'] == 'Generator':
                elm = PlexosGenerator(name=row['name'], category=row['category'])
                self.generators[elm.name] = elm

            elif row['class'] == 'Node':
                elm = PlexosNode(name=row['name'], zone=row['category'])
                self.nodes[elm.name] = elm

            elif row['class'] == 'Line':
                elm = PlexosLine(name=row['name'])
                self.lines[elm.name] = elm
                self.branches[elm.name] = elm

            elif row['class'] == 'Transformer':
                elm = PlexosTransformer(name=row['name'])
                self.transformers[elm.name] = elm
                self.branches[elm.name] = elm

            elif row['class'] == 'Zone':
                elm = PlexosZone(name=row['name'])
                self.zones[elm.name] = elm

            elif row['class'] == 'Region':
                elm = PlexosRegion(name=row['name'])
                self.regions[elm.name] = elm

        # store the branches by type
        self.branches_by_type['Line'] = self.lines
        self.branches_by_type['Transformer'] = self.transformers

        # parse the memberships (node of the elements, region of the nodes, etc...)
        for i, row in memberships.iterrows():
            cls = row['parent_class']
            name = row['parent_object']
            member = row['child_object']

            if cls in ['Line', 'Transformer']:
                if row['collection'] == 'Node From':
                    self.branches_by_type[cls][name].node_from = self.nodes[member]
                elif row['collection'] == 'Node To':
                    self.branches_by_type[cls][name].node_to = self.nodes[member]

            elif cls == 'Node':
                if row['collection'] == 'Region':
                    self.nodes[name].region = self.regions[member]
                elif row['collection'] == 'Zone':
                    self.nodes[name].zone = self.zones[member]

            elif cls == 'Generator':
                if row['collection'] == 'Nodes':
                    self.generators[name].node = self.nodes[member]
                    self.nodes[member].generators.append(self.generators[name])

        # make dictionary of file objects -> file paths
        files_info = properties[properties['child_class'] == 'Data File']
        file_dict = {row['child_object']: row['path'] for i, row in files_info.iterrows()}

        # parse the properties of the objects

        if self.text_func is not None:  # show progress
            self.text_func('Parsing properties')

        used_profiles = set()
        for i, row in properties.iterrows():
            cls = row['child_class']
            name = row['child_object']
            prop = row['property']
            file_obj = row['filename']

            if isinstance(file_obj, float):
                if np.isnan(file_obj):
                    file_obj = None

            if cls == 'Node':

                if prop == 'Voltage':
                    self.nodes[name].voltage = float(row['value'])

                elif prop == 'Fixed Load':
                    self.nodes[name].load = float(row['value'])

                    if file_obj is not None:
                        self.nodes[name].load_prof = file_obj
                        used_profiles.add(file_obj)

            elif cls in ['Line', 'Transformer']:

                if prop == 'Units':
                    self.branches_by_type[cls][name].units = float(row['value'])

                elif prop == 'Resistance':
                    self.branches_by_type[cls][name].r = float(row['value'])

                elif prop == 'Reactance':
                    self.branches_by_type[cls][name].x = float(row['value'])

                elif prop == 'Max Flow':
                    self.branches_by_type[cls][name].rate_max = float(row['value'])

                    if file_obj is not None:
                        self.branches_by_type[cls][name].rate_max_prof = file_obj
                        used_profiles.add(file_obj)

                elif prop == 'Min Flow':
                    self.branches_by_type[cls][name].rate_min = float(row['value'])

                    if file_obj is not None:
                        self.branches_by_type[cls][name].rate_min_prof = file_obj
                        used_profiles.add(file_obj)

            elif cls == 'Generator':

                if prop == 'Max Capacity':
                    self.generators[name].p_max = float(row['value'])

                elif prop == 'Min Stable Level':
                    self.generators[name].p_min = float(row['value'])

                elif prop == 'Rating Factor':
                    if file_obj is not None:
                        self.generators[name].rating_factor_prof = file_obj
                        used_profiles.add(file_obj)

                elif prop == 'Aux Fixed':
                    if file_obj is not None:
                        self.generators[name].rating_factor_prof = file_obj
                        used_profiles.add(file_obj)

                elif prop == 'Must-Run Units':
                    if file_obj is not None:
                        self.generators[name].rating_factor_prof = file_obj
                        used_profiles.add(file_obj)

            if self.prog_func is not None:
                self.prog_func(i + 1 / properties.shape[0] * 100)

        # load the profiles that are used
        if self.load_profiles:
            if self.text_func is not None:
                self.text_func('Loading input profiles')

            for i, file_obj in enumerate(used_profiles):
                self.load_profile_if_necessary(key=file_obj,
                                               path=file_dict[file_obj],
                                               zip_file_pointer=zip_file_pointer)
                # show progress
                if self.prog_func is not None:
                    self.prog_func(i + 1 / len(used_profiles) * 100)
        else:
            if self.text_func is not None:
                self.text_func('Skipping input profiles')

    def get_buses_dictionary(self):
        """
        Get dictionary relating the bus name to the latitude, longitude, voltage and name
        >>> bus_dict[name]
        >>> latitude, longitude, voltage, name
        :return: dictionary
        """
        return {b.name: (b.latitude, b.longitude, b.voltage, b.name) for b in self.nodes.values()}

    def get_all_branches_dictionary(self):
        """
        Returns a dictionary with all the branches by the name
        :return: dictionary name -> object
        """
        z = self.lines.copy()  # start with x's keys and values
        z.update(self.transformers)
        return z

    def get_branch_ratings(self, n=None):
        """
        Get DataFrame with the dynamic branch ratings
        :param n: number of time steps to extrapolate the profile
        :return: DataFrame
        """

        """
        plexos index
        M1-D1-P1
        M4-D1-P1

        M1-3
        M4-5
        M6-8
        M9-10
        M11-12
        """

        # merge the max rating profiles for each of the branches with profiles
        df = None
        for name, branch in self.get_all_branches_dictionary().items():

            if branch.rate_max_prof is not None:
                profile = self.data_profiles[branch.rate_max_prof].set_index('Pattern')
                new_df = profile[[name]]
                if df is None:
                    df = new_df
                else:
                    df = pd.concat([df, new_df], axis=1, sort=False)

        if n is None:
            return df

        # generate dates from the stupid plexos index
        for i, val in enumerate(df.index.values):
            m = val.replace('M', '').split('-')[0]
            df.index.values[i] = datetime(2000, int(m), 1)

        # generate new index according to the number of elements n
        new_index = [datetime(2000, 1, 1) + timedelta(hours=i) for i in range(n)]

        # forward fill...
        df2 = df.reindex(new_index).ffill()

        return df2


def plexos_to_gridcal(mdl: PlexosModel, plexos_results_folder, time_indices=None, text_func=None, prog_func=None):
    """
    Reads plexos model with results and creates a GridCal model
    :param mdl: Plexos model instance
    :param plexos_results_folder: plexos results folder
    :param time_indices: time indices to sample
    :return: MultiCircuit instance
    """
    gen_df = get_st_generation_sent_out(plexos_results_folder=plexos_results_folder)
    load_df = get_st_node_load(plexos_results_folder=plexos_results_folder)
    rating_df = mdl.get_branch_ratings(n=load_df.shape[0])

    circuit = MultiCircuit()
    circuit.name = 'Grid from PLEXOS model'
    circuit.comments = 'Grid from PLEXOS model created by Red Eléctrica de España'
    # keep the time profile

    if 'DATETIME' in gen_df.columns.values:
        t = pd.to_datetime(gen_df['DATETIME']).values
    else:
        t = gen_df.index.values

    if time_indices is not None:
        circuit.time_profile = t[time_indices]
    else:
        circuit.time_profile = t

    n_total = len(mdl.nodes) + len(mdl.generators) + len(mdl.branches)
    nn = 0

    # add the buses and the loads (in plexos there is only one load per bus)
    bus_dict = dict()
    for name, elm in mdl.nodes.items():
        bus = Bus(name=name, vnom=elm.voltage,
                  latitude=elm.longitude,
                  longitude=-elm.latitude,
                  area=elm.region.name,
                  zone=elm.zone.name)

        # add the bus to the buses dictionary
        bus_dict[name] = bus

        # add the bus to the circuit
        bus.ensure_profiles_exist(circuit.time_profile)
        circuit.add_bus(bus)

        # add the load and its profile if it is not zero
        if name in load_df.columns.values:
            if time_indices is not None:
                load_profile = load_df[name].values[time_indices]
            else:
                load_profile = load_df[name].values

            if (load_profile != 0).any():
                load = Load(name='Load@' + name,
                            P=elm.load,
                            Q=elm.load * 0.8,
                            P_prof=load_profile,
                            Q_prof=load_profile * 0.8,)

                load.ensure_profiles_exist(circuit.time_profile)
                bus.add_device(load)

        if text_func is not None:
            text_func("Creating GridCal model: Buses")

        if prog_func is not None:
            prog_func((nn / n_total) * 100)
        nn += 1

    # add the generators
    for name, elm in mdl.generators.items():
        if name in gen_df.columns.values:

            if time_indices is not None:
                gen_profile = gen_df[name].values[time_indices]
            else:
                gen_profile = gen_df[name].values

            gen = Generator(name=name,
                            power_prof=gen_profile,
                            p_min=elm.p_min,
                            p_max=elm.p_max,
                            technology=GeneratorTechnologyType.CombinedCycle)

            bus = bus_dict[elm.node.name]
            gen.ensure_profiles_exist(circuit.time_profile)
            bus.add_device(gen)

        if text_func is not None:
            text_func("Creating GridCal model: Buses")

        if prog_func is not None:
            prog_func((nn / n_total) * 100)
        nn += 1

    # add the lines
    for name, elm in mdl.lines.items():
        bus_f = bus_dict[elm.node_from.name]
        bus_t = bus_dict[elm.node_to.name]

        if name in rating_df.columns.values:
            profile = rating_df[name].values
            rating = profile.max()
        else:
            profile = None
            rating = elm.rate_max

        br = Branch(bus_from=bus_f,
                    bus_to=bus_t,
                    name=name,
                    active=elm.units,
                    r=elm.r,
                    x=elm.x,
                    rate=rating,
                    branch_type=BranchType.Line)
        br.rate_prof = profile
        br.ensure_profiles_exist(circuit.time_profile)
        circuit.add_branch(br)

        if text_func is not None:
            text_func("Creating GridCal model: Buses")

        if prog_func is not None:
            prog_func(nn / n_total * 100)
        nn += 1

    # add the transformers
    for name, elm in mdl.transformers.items():
        bus_f = bus_dict[elm.node_from.name]
        bus_t = bus_dict[elm.node_to.name]

        if name in rating_df.columns.values:
            profile = rating_df[name].values
            rating = profile.max()
        else:
            profile = None
            rating = elm.rate_max

        br = Branch(bus_from=bus_f,
                    bus_to=bus_t,
                    name=name,
                    active=elm.units,
                    r=elm.r,
                    x=elm.x,
                    rate=rating,
                    branch_type=BranchType.Transformer)
        br.rate_prof = profile
        br.ensure_profiles_exist(circuit.time_profile)
        circuit.add_branch(br)

        if text_func is not None:
            text_func("Creating GridCal model: Buses")

        if prog_func is not None:
            prog_func(nn / n_total * 100)
        nn += 1

    return circuit


if __name__ == '__main__':

    folder = ''
    f = os.path.join(folder, 'BASE N.zip')
    folder = os.path.dirname(f)

    mdl_ = PlexosModel(fname=f)

    step = 4
    time_indices_ = np.arange(0, 8760, step)

    circuit_ = plexos_to_gridcal(mdl=mdl_, plexos_results_folder=folder, time_indices=time_indices_)

    fs = FileSave(circuit=circuit_, file_name='spain_plexos(base con restricciones).gridcal', text_func=print)
    fs.save()



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import numpy as np


from GridCal.Engine.Simulations.PowerFlow.time_series_driver import TimeSeriesResults
from GridCal.Engine.Core.time_series_pf_data import TimeCircuit


class TimeSeriesResultsAnalysis:

    def __init__(self, numerical_circuit: TimeCircuit, results: TimeSeriesResults):
        """
        Constructor
        :param numerical_circuit: TimeCircuit instance
        :param results: TimeSeriesResults instance
        """
        self.numerical_circuit = numerical_circuit

        self.res = results

        self.branch_overload_frequency = np.zeros(self.res.m)
        self.bus_under_voltage_frequency = np.zeros(self.res.n)
        self.bus_over_voltage_frequency = np.zeros(self.res.n)

        self.branch_overload_accumulated = np.zeros(self.res.m, dtype=complex)
        self.bus_under_voltage_accumulated = np.zeros(self.res.n, dtype=complex)
        self.bus_over_voltage_accumulated = np.zeros(self.res.n, dtype=complex)

        self.buses_selected_for_storage_frequency = np.zeros(self.res.n)

        self.__run__()

    def __run__(self):
        """
        Run the analysis
        Returns:

        """


        '''
        Optimal storage locations are those nodes where there
        are voltage problems and those nodes receiving the flow
        of current in case of overloads for the time series
        simulation.

            Returns:
        '''

        self.buses_selected_for_storage_frequency = np.zeros(self.res.n)

        rates = self.numerical_circuit.Rates.T

        for t in range(self.res.nt):

            bus_voltage = np.abs(self.res.voltage[t])

            branch_loading = np.abs(self.res.loading[t])

            buses_over = np.where(bus_voltage > self.numerical_circuit.bus_data.Vmax)[0]

            buses_under = np.where(bus_voltage < self.numerical_circuit.bus_data.Vmin)[0]

            branches_over = np.where(branch_loading > 1.0)[0]

            # get the buses from the selected branches
            flow_dir = self.res.Sf[t, branches_over].real

            branches_w_from = np.where(flow_dir > 0)[0]
            branches_w_to = np.where(flow_dir < 0)[0]

            buses_f = self.numerical_circuit.F[branches_w_from]
            buses_t = self.numerical_circuit.T[branches_w_to]

            # branches
            self.branch_overload_frequency[branches_over] += 1
            self.bus_under_voltage_frequency[buses_under] += 1
            self.bus_over_voltage_frequency[buses_over] += 1

            inc_loading = self.res.Sf[t, branches_over] - rates[t, branches_over]
            inc_over = bus_voltage[buses_over] - self.numerical_circuit.bus_data.Vmax[buses_over]
            inc_under = self.numerical_circuit.bus_data.Vmin[buses_under] - bus_voltage[buses_under]

            self.branch_overload_accumulated[branches_over] += inc_loading
            self.bus_under_voltage_accumulated[buses_under] += inc_under
            self.bus_over_voltage_accumulated[buses_over] += inc_over

            # buses for storage
            self.buses_selected_for_storage_frequency[buses_over] += 1
            self.buses_selected_for_storage_frequency[buses_under] += 1
            self.buses_selected_for_storage_frequency[buses_f] += 1
            self.buses_selected_for_storage_frequency[buses_t] += 1



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import numpy as np
import scipy.sparse as sp
import GridCal.Engine.Core.topology as tp


class ShuntData:

    def __init__(self, nshunt, nbus, ntime=1):
        """

        :param nshunt:
        :param nbus:
        """
        self.nshunt = nshunt
        self.ntime = ntime

        self.shunt_names = np.empty(nshunt, dtype=object)

        self.shunt_active = np.zeros((nshunt, ntime), dtype=bool)
        self.shunt_admittance = np.zeros((nshunt, ntime), dtype=complex)

        self.shunt_controlled = np.zeros(nshunt, dtype=bool)
        self.shunt_b_min = np.zeros(nshunt, dtype=float)
        self.shunt_b_max = np.zeros(nshunt, dtype=float)

        self.C_bus_shunt = sp.lil_matrix((nbus, nshunt), dtype=int)

    def slice(self, elm_idx, bus_idx, time_idx=None):
        """

        :param elm_idx:
        :param bus_idx:
        :param time_idx:
        :return:
        """

        if time_idx is None:
            tidx = elm_idx
        else:
            tidx = np.ix_(elm_idx, time_idx)

        data = ShuntData(nshunt=len(elm_idx), nbus=len(bus_idx))

        data.shunt_names = self.shunt_names[elm_idx]

        data.shunt_controlled = self.shunt_controlled[elm_idx]
        data.shunt_b_min = self.shunt_b_min[elm_idx]
        data.shunt_b_max = self.shunt_b_max[elm_idx]

        data.shunt_active = self.shunt_active[tidx]
        data.shunt_admittance = self.shunt_admittance[tidx]

        data.C_bus_shunt = self.C_bus_shunt[np.ix_(bus_idx, elm_idx)]

        return data

    def get_island(self, bus_idx, t_idx=0):
        if self.nshunt:
            return tp.get_elements_of_the_island(self.C_bus_shunt.T, bus_idx, active=self.shunt_active[t_idx])
        else:
            return np.zeros(0, dtype=int)

    def get_controlled_per_bus(self):
        return self.C_bus_shunt * (self.shunt_controlled * self.shunt_active)

    def get_injections_per_bus(self):
        return self.C_bus_shunt * (self.shunt_admittance * self.shunt_active)

    def get_b_max_per_bus(self):
        return self.C_bus_shunt * (self.shunt_b_max.reshape(-1, 1) * self.shunt_active)

    def get_b_min_per_bus(self):
        return self.C_bus_shunt * (self.shunt_b_min.reshape(-1, 1) * self.shunt_active)

    def __len__(self):
        return self.nshunt




# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

from GridCal.Engine.Devices.editable_device import EditableDevice, DeviceType, GCProp


class Wire(EditableDevice):

    def __init__(self, name='', idtag=None, gmr=0.01, r=0.01, x=0.0, max_current=1):
        """
        Wire definition
        :param name: Name of the wire type
        :param gmr: Geometric Mean Radius (m)
        :param r: Resistance per unit length (Ohm / km)
        :param x: Reactance per unit length (Ohm / km)
        :param max_current: Maximum current of the conductor in (kA)

        """

        EditableDevice.__init__(self,
                                name=name,
                                idtag=idtag,
                                active=True,
                                device_type=DeviceType.WireDevice,
                                editable_headers={'name': GCProp('', str, "Name of the conductor"),
                                                  'idtag': GCProp('', str, 'Unique ID'),
                                                  'r': GCProp('Ohm/km', float, "resistance of the conductor"),
                                                  'x': GCProp('Ohm/km', float, "reactance of the conductor"),
                                                  'gmr': GCProp('m', float, "Geometric Mean Radius of the conductor"),
                                                  'max_current': GCProp('kA', float, "Maximum current of the conductor")
                                                  },
                                non_editable_attributes=list(),
                                properties_with_profile={})

        # self.wire_name = name
        self.r = r
        self.x = x
        self.gmr = gmr
        self.max_current = max_current

    def copy(self):
        """
        Copy of the wire
        :return:
        """
        # name='', idtag=None, gmr=0.01, r=0.01, x=0.0, max_current=1
        return Wire(name=self.name, gmr=self.gmr, r=self.r, x=self.x, max_current=self.max_current)



# Resource object code (Python 3)
# Created by: object code
# Created by: The Resource Compiler for Qt version 5.15.0
# WARNING! All changes made in this file will be lost!

from PySide2 import QtCore

qt_resource_data = b"\
\x00\x00\x07*\
\x00\
\x00& x\x9c\xddZ\xddo\xe3\xb8\x11\x7f\xcf_\xa1\
j_vQ\x8b\xe2\xf0\x9b\xae\x9d\x03z\x8b\x03\xee\xb5\
\xbd\xc3=\x16\x8a\xc4\xd8\xea\xca\x92!\xc9qr\x7f}\
\x87\xb2%\x7f(.\x1c \x05\x5c+Yl\xc8\x19R\
3?\xce'\xed\xd9O\xaf\xab\x22xqu\x93W\xe5\
<\x04B\xc3\xc0\x95i\x95\xe5\xe5b\x1e\xfe\xfe\xdb/\
\x91\x09\x83\xa6M\xca,)\xaa\xd2\xcd\xc3\xb2\x0a\x7fz\
|\x98\xfd%\x8a\x82\x9fk\x97\xb4.\x0b\xb6y\xbb\x0c\
~-\x7f4i\xb2v\xc1\xd7e\xdb\xae\xa7q\xbc\xdd\
nI\xbe\x9f$U\xbd\x88\xbf\x05Q\xf4\xf8\xf00k\
^\x16\x0fA\x10\xe0{\xcbf\x9a\xa5\xf3p\xbf`\xbd\
\xa9\x8b\x8e1KcW\xb8\x95+\xdb&\x06\x02qx\
`O\x0f\xec\xa9\x7f{\xfe\xe2\xd2j\xb5\xaa\xca\xa6[\
Y6_\x8e\x98\xeb\xecy\xe0\xf6\xd2ly\xc7\x04\xd6\
\xda\x98\xb2\x98\xb1\x089\xa2\xe6\xadl\x93\xd7\xe8t)\
\xca\xf8\xdeRF)\x8d\x91v\xe0\xbc\x8ek\xda \xa0\
k\xfc7\xb0\xf7\x13\xa4\xa96u\xea\x9eq\x9d#\xa5\
k\xe3\xef\xbf}\x1f\x88\x11%Y\x9b\x1dm\xd3\xe3y\
\xf2\xd6\x13\x90\xcbd\xe5\x9au\x92\xba&\xee\xe7\xbb\xf5\
\xdb<k\x97\xf3P\x98n\xb4t\xf9b\xd9\x0e\xc3\x97\
\xdcm\xff^\xbd\xceC\x1a\xd0@\x18\xfc%\xd4?\xd0\
Q\xf3l\x1e\xa26l\xc7z\xb0\x94=u\xff\x96\xe9\
@\xa1\xc42\xc2\x82\xaf2\xe5\xce\xd0l\x120\x0a:\
\xa2&\xa2\xea[\xb7\xa4Wo\x9aU\xa9\x97w\x1e&\
\x9b\xb6Z\xe1i\xa6\xff*\x92\xb7j\xd3\x12\x8f\xde#\
\xf2\xce2\xf7\xdc\xf85;1\xfcH\x84A\xdc\x91\x86\
m\xfc\x1e\x99\xd7\xe1\xc0\xf8\x944;\xbd\x83`\x9d,\
\xd0F\x8a\xaa\x9e\x87_\x9e\xbbgOx\xaa\xea\xcc\xd5\
=Iu\xcf\x09\xa9B\x1c\xf3\xf6m\xe7\x15\xfb\xbd{\
m\xfd\xae\x03\x9d\xbeOo\x96IVm\xe7!;'\
\xfeYU+\x04\x9fh\xa3\x844\xfc\x9c\x9c\xe2ID\
x\x06\x96\x82\xb2fD\xc5\x172 \x9aq&FK\
\x11\xd1\x8dw\x9chS\xe6-\x1a\xe7\xfau\xb4|S\
\xd7\x9e\x01\x81v\xa8\xf7\x82\x0b\x01{\x9efYm\x17\
\xb5\x87\xef9)\x06\xfc.\xee\xb4\xcdKT/\xda[\
\x16\x185\x92f\xcf\xd1[\x1bP&/\xb0\xa0\xc6R\
_\xa0y}/\xd1V\xc9k\xbe\xca\xfft(3\xf4\
v\xb1rm\x92%mr\xb0\x86~Fw6\x85,\
\xe8\xf7\xd3\x7f|\xffe7\xc2q\x9aN\xff\xa8\xea\x1f\
\xfb!>\x9e!yBK\x9c\x87\xe1\xe30=\xcb\xd2\
)z*Z\xeac\xbe\xc2\x03\xf6N\xfeW\xf4\xccY\
| \x9c0\xb7okw\xd8t\xb7m\xedv.\xff\
n\xdc\xcb\xd2U\xee\x17\xc5\xffl\xf3\xa2\xf8\xd5\xbfd\
\xaf\xd6\xd1\xa6y[\xb8\xc3\xe4,\xdeK\xbf\xd7->\
Rn\x16\xf7\xaaw\xa3\xc5\x19\x88E\xf2\xe4\x8ay\xf8\
s\xb2N\x028GxQW\x9b\xf5\xaa\xcaP\xd0\xce\
V\xc2\x03\x9e\xdd\xb8_\xd0\xd6I\xd9x\xe5\xe7a\xf7\
g\x819\xe1+\x9dD@\xa9 \x5c1\xf6\xadG}\
\xd1\xab\xe1\xf786\xbc\x93M\x10\xc4:\x7f\xfdJ\x89\
Q\x0a\x14\x97|B\xfd\xcfa\xc8\x88\x92Z\x1bm'\
\xc05\x01t\x04\xf86l\xd4\xb4o\x05J\xfc\x8c\xe8\
M\xf7\xee\xfe\xb7\xa6\xad\xab\x1fn\xfaEd\xfe'<\
\x9cz^\xa7\xc5\xd1\xf9\xd4\xde\xd3ex\x98\xf0\xae\x86\
j\x18\xc2a\xb0\xcen\x1e\xcd\x95\x09\x22\x85\xb4\xc6\x1c\
\xcd{\xbd\xd6I\xbb\xe4\x9c\xab\xa3\xe9\xf7d\xf2\x83h\
\x1f@\xa6\xb0\x1b\xd6\x9b\xc2M\xdd\x8b+\xab,\x1b\x84\
\xb6\xdd\xb3\x1f\xee\x9cm\x0a\xeb\xd7~\xa2\xc8K\x87\xc7\
5}\xda\xb4\xed\xf1\xdc\xbf\xab\xbc\x9c\xa2-\xb9\xba\x9f\
\x1d^vdP\xd7B\x00\x9c\x00\xb7\xe7\x10 \xfa\xd4\
PA\xd9\x05\x08\x22~W 0\x85\xe6\x0c\xea\x0c\x04\
\xc1\x89\x15\x16\xad\xf1\x12\x08\xf7e\x09\xdcz\xef\xe3\xe7\
 P\x22\xa8\xe0\x82_\x02A\xdf\x15\x08B\x13M\xcd\
(\x22\x18\x84\xc6jy\xd1\x1d\xe4}\x81 1$3\
8\x03\x01\x04aRQ\xb8\x1c\x13\xee\x0b\x05.\x09H\
\x90g(HB\xadB\x13\xb9\x94\x1c\x22\xc5\xee\x0a\x05\
\x869Y\x99QT \x06\x0dAh\xb8\x84\x82\xbd+\
\x10\x00A\xb0\xea<?\x00\x10!,\xb0\x8b\xa6\x00\xf7\
\x05\x82%x\xe2\xf6<4R\xc2\x94\x94T^\x04\xe1\
\xce2\x04:\x04g\xf6<,p@\x8f0\xe2$s\
\x9c\xc1@\xef\x0a\x06\xce|\x9f:\x82\x01\x1d\x05\x8b\xff\
\xcb\x1eqg\x81\x01\xebd#\xd8\xb9K \xa7\x06\xa6\
/&J\xb8\xb7\xf2\xd9\x12\x8a\xa9r\x5c/\x08\x90J\
],\x9f\xe1\xff\xa5\x80\xf6\x12\x1f\xe90\xdc\xb7T%\
n\xdbVu\x94n\xea\x97\xa4\xdd\xd4(?}G[\
\xc1N\xc0\xf1\xd7\x16\x81\x22T\x1a-\xad\x9c ~\xd8\
\x83)\xc9\x02\x8d\xa1\x94R\x8c\xa6\x13I\xb8\xb6\xe6\x1a\
t\xae\x87\xe3\xe3`t\x83\x22\xc7\xff\xa6\xa2\x9f\xcb\x92\
f\x99\xd4u\xf26-\xab\xd2\xfd\xafa\xe3\xe7\xb0\x01\
\xd6_\xc6R&<l\x9aX\x90A\x1aP\xc2)\xd5\
 &\x116+\x8c\xa9@x \x01\xd8$\xd2D\x82\
\x10\xa3\x89OE\xd6\xdf\x8a\x1a\x83\xa2\x88[\xb28y\
\x0e\x1d\xf3\x17-\x06\xabW\x84\x0e\x10:\x8e\xb8pB\
\x19\x160|\x12a\x87#\xe8U\xc1\xf9\x03\xb8`\x17\
a\xb9f\xfa\x96p\xd1#\x93\xf2\x99\x8c\xa2\xa4\x93\xae\
\xd0c\x18\xb4\x02\xc03\x15\xda\x084)N\xb0\x15\xba\
\xaa\x92\xfb\x88\xc1\x08eAiuK\xc0\xd8q\x88\xe2\
h#Rz`\xb0\x19\x12\xd2\x06\xd0]\x89p4\xf6\
\x09\x10k\xf5g\x1b\x8c\x15\xfe\xe6\x0d\xf8\x0d\xe1\xa2F\
\xa1\x1b\x1dIjm\xb5w$n\x88VB\x06\x11\xe2\
\xa3\x8c\xe1v\xa2\x08c\xf4sC7:\x12Z\xa5e\
7e/j\x14\x9b1\x1e3\x90\xc6\xb2\xc9\xee\x1e\x05\
sZ\x10\x81\xff<Cj\x0e\x93\x08\x88\xd2W\xf5\x02\
\x1f2\x18.\x05\x03sK\xc0\x8c\x1c\x09\xabD-\xac\
\x06\x9f\xeb\xb1*\xe2T\xeb\x003<gT\xa0\x0d\x01\
\xe1X\x08|6.\x00XA\xdcR\x09\xa4G~$\
\x18\x91\x06\xe3\xcb\xa4\xbb\x80\xa5\x9a\x8b \xb2\xc4sJ\
\xe6\xe3.\x97\xea\xb3\xe3\xae\xb6T\xe0\xdbn\xc9\x5c\xf4\
(Q\x0b\x7f|\xfe\xa6m\x17_$\xd5\x22\xc0\x1ce\
%\xef\xbc\x08\xa3\x01U\x9f\x1e`8U\x1a\xb4\xbd%\
`\xc6\x99\x1a\x83,(e\xba\xc0K\x09\xa0\xe3\xa3\x1f\
\x01\x005\x1a\xe3\xae0\x9f\x0e\x8b\x16\x1c\xbbzzK\
~t\xd2\x16\xec\xe2.\xf6d\x98\xa0YW\xd8\xf9K\
|\xa3\x03\x835\xa9Tz\xc2\x08hyU\xb7\xf9\x11\
T\xb0K\x91\x8c\xdeR\x96\xd6\xa7=\xf5.\xec2b\
0\x1f)\x9f\x8f8\x10\xa1\xb13\xc0\x0e\xc1\x80\xc2\xfc\
\xe4;\x01\x8bP}64\x9cbok\xd9\x8dAs\
\xd2i\xef\xc0\xb1\xd8\x08\xe9]\xb2\xc6 \x83\x15\x1dF\
_\x1f\x85\x95B\xc0\xb0\x9ca n\x1b\x9cY\xbc\xe8\
?\xf3G$\x1e\xde\x91\xb3\xebPO\xee\x10\xf0\xf4\x85\
\xc5\x06P\xf2\xff**'Fj\x81\xf1E\x8e\x84\xaa\
\xabM\xd9\xebyE\x83< \xe8\xcf\xc3\xcb)\xe0\xe8\
\x03\xe4\xfe\xab;\xc2\xc72\xb4\xd4\x810|\x8bgD\
\xf1\x17+\xbe\xdf\xd4\xf4(\xa1\xee>\xcc\xc6\xca\xd3\x1e\
\xb5pu7K\xc0(-\x87oPx\xccf\xfe\xcb\
\x0d\x8f\x0f\xff\x01\x9bj\x14\x0c\
\x00\x00\x0c/\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<!-- Creat\
ed with Inkscape\
 (http://www.ink\
scape.org/) -->\x0a\
\x0a<svg\x0a   xmlns:d\
c=\x22http://purl.o\
rg/dc/elements/1\
.1/\x22\x0a   xmlns:cc\
=\x22http://creativ\
ecommons.org/ns#\
\x22\x0a   xmlns:rdf=\x22\
http://www.w3.or\
g/1999/02/22-rdf\
-syntax-ns#\x22\x0a   \
xmlns:svg=\x22http:\
//www.w3.org/200\
0/svg\x22\x0a   xmlns=\
\x22http://www.w3.o\
rg/2000/svg\x22\x0a   \
xmlns:sodipodi=\x22\
http://sodipodi.\
sourceforge.net/\
DTD/sodipodi-0.d\
td\x22\x0a   xmlns:ink\
scape=\x22http://ww\
w.inkscape.org/n\
amespaces/inksca\
pe\x22\x0a   width=\x2248\
\x22\x0a   height=\x2248\x22\
\x0a   viewBox=\x220 0\
 48.000001 48.00\
0001\x22\x0a   id=\x22svg\
2\x22\x0a   version=\x221\
.1\x22\x0a   inkscape:\
version=\x220.92.2 \
(5c3e80d, 2017-0\
8-06)\x22\x0a   sodipo\
di:docname=\x22resi\
ze.svg\x22>\x0a  <defs\
\x0a     id=\x22defs4\x22\
 />\x0a  <sodipodi:\
namedview\x0a     i\
d=\x22base\x22\x0a     pa\
gecolor=\x22#ffffff\
\x22\x0a     bordercol\
or=\x22#666666\x22\x0a   \
  borderopacity=\
\x221.0\x22\x0a     inksc\
ape:pageopacity=\
\x220.0\x22\x0a     inksc\
ape:pageshadow=\x22\
2\x22\x0a     inkscape\
:zoom=\x227.9195959\
\x22\x0a     inkscape:\
cx=\x2223.580795\x22\x0a \
    inkscape:cy=\
\x2231.52398\x22\x0a     \
inkscape:documen\
t-units=\x22px\x22\x0a   \
  inkscape:curre\
nt-layer=\x22layer1\
\x22\x0a     showgrid=\
\x22false\x22\x0a     uni\
ts=\x22px\x22\x0a     ink\
scape:window-wid\
th=\x221863\x22\x0a     i\
nkscape:window-h\
eight=\x221025\x22\x0a   \
  inkscape:windo\
w-x=\x2257\x22\x0a     in\
kscape:window-y=\
\x2227\x22\x0a     inksca\
pe:window-maximi\
zed=\x221\x22 />\x0a  <me\
tadata\x0a     id=\x22\
metadata7\x22>\x0a    \
<rdf:RDF>\x0a      \
<cc:Work\x0a       \
  rdf:about=\x22\x22>\x0a\
        <dc:form\
at>image/svg+xml\
</dc:format>\x0a   \
     <dc:type\x0a  \
         rdf:res\
ource=\x22http://pu\
rl.org/dc/dcmity\
pe/StillImage\x22 /\
>\x0a        <dc:ti\
tle />\x0a      </c\
c:Work>\x0a    </rd\
f:RDF>\x0a  </metad\
ata>\x0a  <g\x0a     i\
nkscape:label=\x22C\
apa 1\x22\x0a     inks\
cape:groupmode=\x22\
layer\x22\x0a     id=\x22\
layer1\x22\x0a     tra\
nsform=\x22translat\
e(0,-1004.3622)\x22\
>\x0a    <rect\x0a    \
   style=\x22fill:n\
one;fill-opacity\
:0.84951453;stro\
ke:#999999;strok\
e-width:3.857414\
25;stroke-linejo\
in:round;stroke-\
miterlimit:4;str\
oke-dasharray:no\
ne\x22\x0a       id=\x22r\
ect4136\x22\x0a       \
width=\x2244.142586\
\x22\x0a       height=\
\x2244.142586\x22\x0a    \
   x=\x221.9287071\x22\
\x0a       y=\x221006.\
2909\x22\x0a       ry=\
\x2210.186751\x22 />\x0a \
   <g\x0a       id=\
\x22g4157\x22\x0a       t\
ransform=\x22matrix\
(0.84603094,0,0,\
0.78857975,5.043\
4872,217.0903)\x22\x0a\
       style=\x22st\
roke:#999999;fil\
l:#999999\x22>\x0a    \
  <path\x0a        \
 sodipodi:nodety\
pes=\x22ccc\x22\x0a      \
   inkscape:conn\
ector-curvature=\
\x220\x22\x0a         id=\
\x22path4138\x22\x0a     \
    d=\x22m 9.18824\
93,1043.1259 28.\
1788757,-28.9184\
 0,0\x22\x0a         s\
tyle=\x22fill:#9999\
99;fill-rule:eve\
nodd;stroke:#999\
999;stroke-width\
:4.74617958;stro\
ke-linecap:butt;\
stroke-linejoin:\
miter;stroke-mit\
erlimit:4;stroke\
-dasharray:none;\
stroke-opacity:1\
\x22 />\x0a      <path\
\x0a         sodipo\
di:nodetypes=\x22cc\
cc\x22\x0a         ink\
scape:connector-\
curvature=\x220\x22\x0a  \
       id=\x22path4\
140-3\x22\x0a         \
d=\x22m 38.871283,1\
019.5549 -6.5631\
79,-7.2978 8.297\
448,-1.8083 z\x22\x0a \
        style=\x22f\
ill:#999999;fill\
-rule:evenodd;st\
roke:#999999;str\
oke-width:1px;st\
roke-linecap:but\
t;stroke-linejoi\
n:miter;stroke-o\
pacity:1\x22 />\x0a   \
   <path\x0a       \
  sodipodi:nodet\
ypes=\x22cccc\x22\x0a    \
     inkscape:co\
nnector-curvatur\
e=\x220\x22\x0a         i\
d=\x22path4140-3-6\x22\
\x0a         d=\x22m 6\
.8693068,1038.42\
58 6.5632192,7.2\
979 -8.2974502,1\
.8083 z\x22\x0a       \
  style=\x22fill:#9\
99999;fill-rule:\
evenodd;stroke:#\
999999;stroke-wi\
dth:1px;stroke-l\
inecap:butt;stro\
ke-linejoin:mite\
r;stroke-opacity\
:1\x22 />\x0a    </g>\x0a\
  </g>\x0a</svg>\x0a\
\x00\x00\x0a\xd3\
\x00\
\x000Hx\x9c\xedZ\xdb\x8e\xdbF\x12}\xf7Wp\
\x95\x17\x1b\xcbn\xf5\xfd\x22\x8f\x1c`\x13\x04\xc8\xebn\
\x16\xfb\xb8\xa0%zFk\x89\x1cH\x1c\x8f\xed\xaf\xcf\
\xa9\xe6E\x12)a\x1dg\x03\xec\x8c\xa3\x81-\xb2\xbb\
\xba\xbb\xfat\xd5\xa9\xaa\x9e\xb9\xf9\xfe\xe3n\x9b}(\
\xf7\x87M]-g\x92\x8bYVV\xabz\xbd\xa9n\
\x97\xb3\x7f\xfe\xf2\x13\x0b\xb3\xec\xd0\x14\xd5\xba\xd8\xd6U\
\xb9\x9cU\xf5\xec\xfb7/n\xfe\xc2X\xf6\xc3\xbe,\
\x9ar\x9d=n\x9a\xbb\xec\xe7\xea\xfdaU\xdc\x97\xd9\
\xcb\xbb\xa6\xb9_\xcc\xe7\x8f\x8f\x8f|\xd35\xf2z\x7f\
;\x7f\x951\xf6\xe6\xc5\x8b\x9b\xc3\x87\xdb\x17Y\x96a\
\xdd\xea\xb0X\xaf\x96\xb3n\xc0\xfd\xc3~\x9b\x04\xd7\xab\
y\xb9-we\xd5\x1c\xe6\x92\xcb\xf9\xec(\xbe:\x8a\
\xafh\xf5\xcd\x87rU\xefvuuH#\xab\xc3w\
'\xc2\xfb\xf5\xbbA\x9a\xb4y\xd4IH\xc6\x18\xe7B\
\xcd\x95b\x90`\x87OUS|d\xe7C\xa1\xe3\xa5\
\xa1J\x081G\xdfQ\xf2\xcb\xa4\x16\x07\x00z\x8f\x7f\
\x83x\xdf\xc0\x0f\xf5\xc3~U\xbe\xc3\xb8\x92We3\
\xff\xf1\x97\x1f\x87N&\xf8\xbaY\x9fL\xd3\xe3y\xb6\
\xea\x19\xc8U\xb1+\x0f\xf7\xc5\xaa<\xcc\xfb\xf64\xfe\
q\xb3n\xee\x963\x13\xd2\xdb]\xb9\xb9\xbdk\x86\xd7\
\x0f\x9b\xf2\xf1o\xf5\xc7\xe5Ld\x223\x81\x0b\xfa\xc8\
\xe3S\x12\xda\xac\x973lJ\xb5#\x8e\x06\xd3\xf5v\
\x8b-\x86\x1e\xc1\xa3\xe2:{\xa9\x8c\xb0\xd6\xb8<S\
B\x06&4\x93\xf2U\x1a\xd2\xefr\xb1\xaeW\xa4\xf6\
rv\xbb\xdf\xac\xff\xbdY\xd5\x15'\xf4\xde@\xe8f\
]\xbe;\x90p\xbb>\xbd\x99Y6O]\xc3x\x1a\
\xbc\xa6=\x1c\x05\xdf\x16\x87v\xdfYv_\xdc\xc2F\
\xb6\xf5~9\xfb\xee]\xfat\x1do\xeb\xfd\xba\xdc\xf7\
].}\xce\xbaj\xe0\xb8i>\xb5^\xd1\xcd\xddo\
\x93f\x1d\xfa\xc5\xe5\xfe\xc3]\xb1\xae\x1f\x9735\xee\
\xfc\x5c\xd7;\x9a\xd5\x1a\x15]\x08\xe3\xee\x15N\x82I\
c\xb9\xb0\xd2\xcaI/)$%\xd7\xd1y3\xee\x04\
\x94\x0f\xe48\xec\xa1\xda40\xce\xfb\x8f\x93\xe1\x0f\xfb\
=\x09l\x8bO%\xf6\x9d\xbe\xfa5\x0ew\xf5#\x9d\
\xc1r\xf6\xae\xd8\x0e\x00^\x9d\xeaqSa\x7f\xac3\
-\x19\xa6\xbav\x12\xbd\xb9I\xa1\xec\x15\x11l\xd9\xc5\
+}\xd8\xb0\xf2W\xfav\xc5\xc7\xcdn\xf3\xb9\x84\xce\
\xb27\x8c]\xd9\x14\xeb\xa2)\x8e\xe6\xd0\xb7\xf8dT\
\x10\x81\xe3/\xfe\xfe\xe3O\xed\x1b\xdeW\xab\xc5\xbf\xea\
\xfd\xfb\xee\x15\x1f\x12(\xde\xd6\x0f\xd0z\xf6fh\xbe\
Y\xaf\x16p\xd5]\xd1\xbc\xd9\xecp\xc2\xe4\xe5\x7f\x85\
k\xde\xcc\x8f\x1dg\xc2\xcd\xa7\xfb\xf28i;\xed\xbe\
l}\xfe\x22\xf1\xadW\xbb\x0d\x0d\x9a\xff\xa3\xd9l\xb7\
?\xd3\x22\xdd\xb6N&\xdd4\xdb\xf2\xd8x3\xef\xb4\
\xef\xf66?\xd9\xdc\xcd\xbc\xdfzz\xbb\x1d\x81\xb8-\
\xde\x96\xdb\xe5\xec\x87\xe2\xbe\xc8&\x87w\xbb\xaf\x1f\xee\
w\xf5\xba\xec\xacdv\xc4\xf3\xccj\x9a}Q\x1dh\
\xf3\xcbYz\xdc\x22(\xbc\x149\x93B\x18\xae\x9dR\
\xafz\xd4o\xfbm\xd0\x1c\xb7\xc1\xebY\xdfpq\x0e\
\x16\xe0\x02Rh\x90\x073\x8a\x07\x13Bx5\x0c9\
4\x9f\xb6P\xed\xd0\xec\xeb\xf7\xe5\xe2\xbb\x98>\xaf\xdb\
W\xd6\xb9\xe6B\x0e\x87wSn\xb7\x9b\xfb\xc3\xc9i\
\xecaW\x86\xc7\xe0\x9d\x0a\xb3\x93f\x98\xa2\xe5\xd2H\
\x11\xb5?iO~'\xa4\xe2\xd2\xc6\xb3v\xc8G\x10\
epZ\x9d\xb6\xd3\x1e\xef\x8b\xe6\xceH\xedN\x9a;\
\xb5\x07\x05_\xbf\xc39\x0f\xea\xd3\x0b;\xefc\xfb\x87\
m\xb9(?\x94U\xbd^\xbf\xbe\xbc\xdd\xe4\x82\x0b\xdb\
\xbfn7U\xf9\x9fzS-p\x82U?\x86\xc1\xb0\
\xca\xfd\x16\xde\xd2,L\xdf\xb6.@R\xfb}\xf1i\
Q!\xbaO\xd1;\xb1\xb2+\xf8\x99 \x1d\x90\x1a\x01\
h\xb8\x93\xd1\x0ak&\x00\x1a\xc1\xa5\xf2z\x04 \xb5\
Jo\x8c\xbe\x02 \x93\xcf\x16B\xcbM\x8cZ\x08;\
\xb1A\xe7@\xa9b\x0a!\xc8\xdf\xd8\xe0F\x10\xea\xc8\
\x9d\x95J\xb9k\x10>_+t\x5c{g\x83\x8f#\
\x08\x1d\xb7^\xf9(\xe5\xd4\x0a\x15\xb7R\x8f\xad\xd0s\
g\x9c\xf0\xd7\xadP=[\x08-\xf7\xc6*sf=\
\xad\x15F\x85\xc4-\xda\x09\x84\xca\x03u\x13G\x10*\
\xcb\xc1\xd2\xc1]\xb5B\xf1\x1c \xa4\xdd\x9clcH\
D\x11,)x#[Z\xadN\x01\x18\x92\xaf\xba\xaa\
\xcaUS\xef\x19\xd2\xb0\x0fE\xf3\xb0\xc7\xf6\xc5E\xa8\
\xfc\xa9\xd1R\x0a\x93\xc1\xed]P\x22\x9a\x1c11\x82\
/\x83\xcc\xa4F\x92\xa8\x90&\xe6,r\xe3\xc5\x05\x03\
\x9db\xfa\xc5 \x9aS\x10\xb1\x81s\x0c\x07d\x13\x98\
O\x08Y=AV\xc0h\x95%`\x01(\xf2jE\
\xc0\xba\xa8\xa5\x8a9\x22\xbe\x07\xbb\xfc\x09\xec\x7f\x07\xd6\
\x8e\x81\xd5\x81k\xe1\xb4V-\xb2\xce{\x991\xaa]\
\xb4\xb4V\x12\xb4\xd6\xc9\x0bq\xe9Oh\xc7\xd0\xfa1\
\xb4\xa0_\x85<J\xc2@\x85\x8a<\x22\xd9&\x86\xd0\
>h\xe3s\xd0\x83\x0a\xf1I\x03{3\xbf\xbd\x5c6\
\x84#\x16\x17\xcb\x06$\xee\xc8\xea\x9d\xd1y@\xf9\x10\
\x831\xaffW\xce\xea\xb7\x1fEP\x13^V\x8a\xe3\
 \xa2\x0ay*\x0f\xac\x886\xd3\xa8\x12\x8c\xb26\xd7\
 \x0f\xa53\xc9\x83\x8e1\xc8\x1cqV\x18\x93y\x14\
\x0c\xd1X\x9f3\x0c\xf0\xeeBv\x9bN*!t\xfd\
\x98\x82\xa0\x9f\xf3c\xd2_vLO*\x1a\x06\x15\xbf\
\x04t\xe6Q\xb6\x19\x11\x03\xcc\xdfDq\xcd\xfc\x9f\x22\
\xa8_\x81\x99\x9e\x1a*(\xc3\xa1\x5c\xa38\xa7$\xf9\
\x88\xcaP}h\x85\xc25\xf7<\x8222\x86\x8aX\
8'\x03,U\xb8\xb3\xe2\xe3[\xc4p\x92*\x5c\xb2\
;\xc9\xa5\x8eB\xc3\x9958Y\xe8g\x84\xd9eg\
\xfe\xbd\xee\xac'\x89\x82\xb4\x08X1h\xb2LE\xb7\
\xce&\xa3\xdc\xc0[\x13a\x99A)\x93\xad2&\x10\
\xdeT\xf0\x1e\xac\xaa\xb5\xf7\x99\x80\x89Z\x97+\x0e{\
uH,('v)\xd9\x00\xf1\xebl\x0b\x1a\xb6\xc2\
\xa0\x04\x841[%\xae\xe5pO\xf1`\xbe\x06\xf5I\
\x0e\xa1\x91\xda\x02\xea(\xdb\x1c\x22\xc4\x989\xb2`\x84\
\xd8\xdcpa\x82&\x04\xa3\xa2\x8c\xc2q\xe1\xd57\x8e\
\xa0\x99\xd2\x81\xe7>\xea\xe0\xc9p5\xa8\xd3F\x03\xab\
D&\xac4\x19\x9dQ\xe19A\xf6G\x84v3\xe5\
\x02\xc1\xe9*U%\xab\x0c\x88O\xc0\x949n\xa4\xf1\
\x08\xea@\xd5\x89\xf8\x8d\xc7v3-\x07\x92\xb1\xc9\xc4\
\xa0\x12N\x8cG\xc4\xf6\x80d\x93\xee\x06\x14G\xc8B\
QK\xd4\xaa]@\xcd\xc0\xbd\x09\xd7.\x0b\xbe\x15\x10\
'I%<X\xa1\xd6O\x96'\x91C\xe2\xff\x8c\xaa\
\xd6\x18m\x00!Z\x11Lf8\x8a.\x1dL\x0e7\
GN\xff\x8c \xfccb\xbd\xfd\x92X\x8f\xac\xd3D\
\x8f\xc2\x0da\x06\xfe\x9e\xc1`\xa54\xa8\xdc(/\x05\
T\x92$\x12\xee\xd2!\xb9\x87\x85\xfbg\x04\xfc\xd7\xa0\
z\xe9> \xba(\xc9t\x0d\xf2\xa4\xe0#\xf2R\xa5\
\x9c\xb3\x0au\xb0\x0c\xe1\xda5\xcbSD\xec\x8a\xa9\xfe\
\xceH\xe4\xd9$\xbe\x1bNe\xbb\xb2\xed\xfd\x95\xb4\xda\
fd\xaa\xd6\x06\xe3\x10\x89\x82u\x81R&m\x8c\x12\
\x06\x89h\x8c\xe2\x09\xd7\xf2W\xaf\x5c\xfc\xf1\x17\xa4W\
/\x92\x8e7\xf5\x97o\x91\x8e\xcb]\xb8\xbd\xd9\x15\xcd\
~\xf3\xf1\xa5@\xce\x1fMD\xfa)\xe8gx\x83q\
;k\x82P\xb9\xa1\x18f\x00\xff\xf5;\x9d\xee<\xd5\
\x84w\x1cG=\xe1\x14\xd5\x07QrcP\x22\xac\xa8\
j\x80\x87\x08\xa2\x1c\xc4E,\x09\xf6\xd1<\x82\x9d\xb8\
5(!\x98DP\xb0\xf4g0:g\x81k\x891\
\xf4\xcb\xca\x1c\xd5\x88\xb1\xf4\x18\xe81\xaa\xee\x11\x13g\
\x22u\x06*X\x84O\xbd\x96\x9e\xfb\xeem\x869\xa5\
\x18\xe6T\xa4\x87$QIU\x0fVFa\x8e\xfc;\
\xb4O(v$\x14t\x14\xc6]\xa4oG+Ps\
\x96\x9a\xa8\x81\xa5f\xd6\xc9\xb2T\x14E\xfa&\xd9\xa4\
L\xd2%\xa9\x924i\x15a'\x9at\x8a\xb0\xa4\x09\
Ks[R\x18*\xb0\xa4M\xf7\x8c5\x04\x1d\xac\xa7\
\x85,oO\xd9w\x0b\x1e\xdfi\xe1\xa4cR\xd1e\
\xa7=\xc06`\xb4\x06\x93\x13<\xfdv\xdb\x05\x12<\
=\xe4\x1d\xe2\x84\x1emC\xa5'\x87G*\xf8\xd2\xa3\
$\x09Z\x0a\x09\x0eU\x82\xe4\x96\x9ar\x1f\xdd>Q\
\x07\xd6J\x0d\x04\x97\xec\x1eS\xa7hk\xf6\x84\xa3\xec\
\xda\xd8I\xff\xe7\xaf\xa2\x94\xff\x81\x8f\x5c\x0f\x16\xbdy\
O\xc9\x8a![\xb62\x86@t\xe5\x02\xf7.\x84d\
\xadQXA@)%\x15E\x05\xe3\x02\x0c\x18[\xf6\
\x9a\xde)\xcd\xb6T\xf0yCE3\xc5g'\x95\xa6\
`\x8cS\xb0i\x0a\x15\x9d@\x0b\x95\xe0\xd6Q\x8d\xe3\
\xb4\x8c6\x1d^Tt\x05\x22P\xf4\xb4\x9e\xe0\x09W\
D \xeb4\xbdcn2\xbe\x80:\xb3u\x06+\x90\
\xd0C\x0bL`\xda\x19BL\xa6\xc81\xd2#\x07`\
\x94\xc0\xaa`Zc4FX\x1c3\xcd\xe4L:)\
\x94\x01R%k\x11F\x072H\xafS\xcdJy\xae\
\xc4\x96\xd0\x131\x91h\x0f\xdcF\x9d\xe6\xa1\x8c\x22\xa7\
BB\xa9\x16\x05\xef\x91\xc6\x91O\xc7\xd6JQ\xe9\x92\
\xca\xed\xe0,\x09J\x0d\xe0\xda\xc16K\x92B\xab\x04\
\xa5\xf1!\xeb\x96\x81}\xb5\xa3\xb3v\x15\x9a\xb6\x1dM\
\xeb\xa6ul\x9e\x06\xa7+\x0c\x02\x86\x8c9\x01\xc3\x06\
dDw\x899 \x93\x80\x91t =0-.-\
[\xb4\xb0\x90T\x82\xa5G\x85xKP\x0a\xd5\x81\x12\
\xf2\x84\x89\x1et\xd5.\xef0iw%e\x1c)K\
\xbfM\xec\xb4\xedA!NJ\xa0\xb0\x1e\x15\x99w\xa0\
\xb0\x0e\x15l\x89u\xb0\x90\x11\x18\xe1\xa3J-\x8e\xca\
7|G#]\xa0\x03\xf1\xc1\xa7\x16\xa9\xf0D\x7f\xda\
\x83Q0!\xb2\x80\xd6\xf2PM\xb3dz\xba\xf5|\
\xe4\xe4!\x10bZS\xaeH\x1cm\x9d#\xc4\x9c\x11\
\x89\x06\xa2\xf52\xb1w\x14\xc4\x0b \xae\xe0\xa2\xa62\
'J\xafIk\xd4\xe31\xb1\x9e\xf6&)--j\
\xf4dDp\x8a\xa4\x0f\x89\x02\x8c4\xdaS\x03\x89\xca\
d2\x90\xccZA\x9cu;Xg\xc3*\xdd\xd8l\
X\xa5\x1b\x9bu\xab\xf8\xbc[\xe4\xff\x90M~s\x9a\
\x04\xde9\xfbK\x92\x81y\x84\x93\x8eB\x96\x88\x08\xd6\
\xda\xbb\x13\xe6ac\xeaa#\xeea\x17\xc8\x87M\xd9\
gD>#\xee\x19Q\xcf9\xf3\xb8\x90\x9f3\x8f\x09\
\xf9\x98y\xe0,\x97\x98gB<\x13\xde\x99\xd0\xce\x98\
u\xc6\xa43\xe6\x9cs\xca\x19|\xabw\xad\xc1\xb3z\
'\xeex\xedH8\x1d\xaf\x9d\x12\x0e\x1b3\x0e\x9bP\
\x0e\x9bp\x0e\x9b\x92\x0e\x9b\xb0\x8e\x1f\xc8x\xc4\xc5'\
\xac\xc3F\xb4\x83\xec\x98\x9d\xf3\x8e\x1a\xebl\xe5\xa0s\
\x07\x8c\x19\xc8\x98\x0d\xd0\xb0\x11\xef\x98|\xc2:\x13\xd2\
\x99p\xce%\xca\x990N\x0av\xa7\x8c\x93\x82\xd0\x19\
\xe5\xb4\x01\xef\x9cs\xd8\x84t\xd8\x84u&\xa43\xe6\
\x9c\x11\xe5\x8c\x19gL8\xe7|\xe3\x06V\xfb\xfc\xc7\
\xd0HW'\xa4\xaf\x1b\xfa\x1b\xd87/~\x05,n\
NV\
\x00\x00\x09,\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<!-- Creat\
ed with Inkscape\
 (http://www.ink\
scape.org/) -->\x0a\
\x0a<svg\x0a   xmlns:d\
c=\x22http://purl.o\
rg/dc/elements/1\
.1/\x22\x0a   xmlns:cc\
=\x22http://creativ\
ecommons.org/ns#\
\x22\x0a   xmlns:rdf=\x22\
http://www.w3.or\
g/1999/02/22-rdf\
-syntax-ns#\x22\x0a   \
xmlns:svg=\x22http:\
//www.w3.org/200\
0/svg\x22\x0a   xmlns=\
\x22http://www.w3.o\
rg/2000/svg\x22\x0a   \
xmlns:sodipodi=\x22\
http://sodipodi.\
sourceforge.net/\
DTD/sodipodi-0.d\
td\x22\x0a   xmlns:ink\
scape=\x22http://ww\
w.inkscape.org/n\
amespaces/inksca\
pe\x22\x0a   version=\x22\
1.1\x22\x0a   id=\x22svg2\
\x22\x0a   width=\x22192\x22\
\x0a   height=\x22192\x22\
\x0a   viewBox=\x220 0\
 192 192\x22\x0a   sod\
ipodi:docname=\x22p\
lus (gray).svg\x22\x0a\
   inkscape:vers\
ion=\x220.92.4 (unk\
nown)\x22>\x0a  <metad\
ata\x0a     id=\x22met\
adata8\x22>\x0a    <rd\
f:RDF>\x0a      <cc\
:Work\x0a         r\
df:about=\x22\x22>\x0a   \
     <dc:format>\
image/svg+xml</d\
c:format>\x0a      \
  <dc:type\x0a     \
      rdf:resour\
ce=\x22http://purl.\
org/dc/dcmitype/\
StillImage\x22 />\x0a \
       <dc:title\
 />\x0a      </cc:W\
ork>\x0a    </rdf:R\
DF>\x0a  </metadata\
>\x0a  <defs\x0a     i\
d=\x22defs6\x22 />\x0a  <\
sodipodi:namedvi\
ew\x0a     pagecolo\
r=\x22#ffffff\x22\x0a    \
 bordercolor=\x22#6\
66666\x22\x0a     bord\
eropacity=\x221\x22\x0a  \
   objecttoleran\
ce=\x2210\x22\x0a     gri\
dtolerance=\x2210\x22\x0a\
     guidetolera\
nce=\x2210\x22\x0a     in\
kscape:pageopaci\
ty=\x220\x22\x0a     inks\
cape:pageshadow=\
\x222\x22\x0a     inkscap\
e:window-width=\x22\
1863\x22\x0a     inksc\
ape:window-heigh\
t=\x221025\x22\x0a     id\
=\x22namedview4\x22\x0a  \
   showgrid=\x22fal\
se\x22\x0a     inkscap\
e:zoom=\x221.738304\
2\x22\x0a     inkscape\
:cx=\x22133.79315\x22\x0a\
     inkscape:cy\
=\x2271.379447\x22\x0a   \
  inkscape:windo\
w-x=\x2257\x22\x0a     in\
kscape:window-y=\
\x2227\x22\x0a     inksca\
pe:window-maximi\
zed=\x221\x22\x0a     ink\
scape:current-la\
yer=\x22svg2\x22 />\x0a  \
<g\x0a     id=\x22g831\
\x22\x0a     style=\x22st\
roke:#aeaeae;str\
oke-linecap:roun\
d;stroke-opacity\
:1;fill:#aeaeae;\
fill-opacity:1\x22>\
\x0a    <path\x0a     \
  sodipodi:nodet\
ypes=\x22cc\x22\x0a      \
 inkscape:connec\
tor-curvature=\x220\
\x22\x0a       id=\x22pat\
h812\x22\x0a       d=\x22\
m 95.091973,56.1\
72243 0.09447,79\
.795097\x22\x0a       \
style=\x22fill:#aea\
eae;fill-rule:ev\
enodd;stroke:#ae\
aeae;stroke-widt\
h:16;stroke-line\
cap:round;stroke\
-linejoin:miter;\
stroke-miterlimi\
t:4;stroke-dasha\
rray:none;stroke\
-opacity:1;fill-\
opacity:1\x22 />\x0a  \
  <path\x0a       s\
odipodi:nodetype\
s=\x22cc\x22\x0a       in\
kscape:connector\
-curvature=\x220\x22\x0a \
      id=\x22path81\
2-3\x22\x0a       d=\x22m\
 134.96147,95.59\
4487 -79.795108,\
0.0945\x22\x0a       s\
tyle=\x22fill:#aeae\
ae;fill-rule:eve\
nodd;stroke:#aea\
eae;stroke-width\
:16;stroke-linec\
ap:round;stroke-\
linejoin:miter;s\
troke-miterlimit\
:4;stroke-dashar\
ray:none;stroke-\
opacity:1;fill-o\
pacity:1\x22 />\x0a  <\
/g>\x0a</svg>\x0a\
\x00\x00\x07\x95\
<\
?xml version=\x221.\
0\x22 encoding=\x22UTF\
-8\x22 standalone=\x22\
no\x22?>\x0a<!-- Creat\
ed with Inkscape\
 (http://www.ink\
scape.org/) -->\x0a\
\x0a<svg\x0a   xmlns:d\
c=\x22http://purl.o\
rg/dc/elements/1\
.1/\x22\x0a   xmlns:cc\
=\x22http://creativ\
ecommons.org/ns#\
\x22\x0a   xmlns:rdf=\x22\
http://www.w3.or\
g/1999/02/22-rdf\
-syntax-ns#\x22\x0a   \
xmlns:svg=\x22http:\
//www.w3.org/200\
0/svg\x22\x0a   xmlns=\
\x22http://www.w3.o\
rg/2000/svg\x22\x0a   \
xmlns:sodipodi=\x22\
http://sodipodi.\
sourceforge.net/\
DTD/sodipodi-0.d\
td\x22\x0a   xmlns:ink\
scape=\x22http://ww\
w.inkscape.org/n\
amespaces/inksca\
pe\x22\x0a   version=\x22\
1.1\x22\x0a   id=\x22svg2\
\x22\x0a   width=\x22192\x22\
\x0a   height=\x22192\x22\
\x0a   viewBox=\x220 0\
 192 192\x22\x0a   sod\
ipodi:docname=\x22m\
inus (gray).svg\x22\
\x0a   inkscape:ver\
sion=\x220.92.4 (un\
known)\x22>\x0a  <meta\
data\x0a     id=\x22me\
tadata8\x22>\x0a    <r\
df:RDF>\x0a      <c\
c:Work\x0a         \
rdf:about=\x22\x22>\x0a  \
      <dc:format\
>image/svg+xml</\
dc:format>\x0a     \
   <dc:type\x0a    \
       rdf:resou\
rce=\x22http://purl\
.org/dc/dcmitype\
/StillImage\x22 />\x0a\
        <dc:titl\
e />\x0a      </cc:\
Work>\x0a    </rdf:\
RDF>\x0a  </metadat\
a>\x0a  <defs\x0a     \
id=\x22defs6\x22 />\x0a  \
<sodipodi:namedv\
iew\x0a     pagecol\
or=\x22#ffffff\x22\x0a   \
  bordercolor=\x22#\
666666\x22\x0a     bor\
deropacity=\x221\x22\x0a \
    objecttolera\
nce=\x2210\x22\x0a     gr\
idtolerance=\x2210\x22\
\x0a     guidetoler\
ance=\x2210\x22\x0a     i\
nkscape:pageopac\
ity=\x220\x22\x0a     ink\
scape:pageshadow\
=\x222\x22\x0a     inksca\
pe:window-width=\
\x221863\x22\x0a     inks\
cape:window-heig\
ht=\x221025\x22\x0a     i\
d=\x22namedview4\x22\x0a \
    showgrid=\x22fa\
lse\x22\x0a     inksca\
pe:zoom=\x221.73830\
42\x22\x0a     inkscap\
e:cx=\x22-319.39807\
\x22\x0a     inkscape:\
cy=\x22209.44504\x22\x0a \
    inkscape:win\
dow-x=\x2257\x22\x0a     \
inkscape:window-\
y=\x2227\x22\x0a     inks\
cape:window-maxi\
mized=\x221\x22\x0a     i\
nkscape:current-\
layer=\x22g831\x22 />\x0a\
  <g\x0a     id=\x22g8\
31\x22\x0a     style=\x22\
stroke:#b3b3b3;s\
troke-linecap:ro\
und\x22>\x0a    <path\x0a\
       sodipodi:\
nodetypes=\x22cc\x22\x0a \
      inkscape:c\
onnector-curvatu\
re=\x220\x22\x0a       id\
=\x22path812-3\x22\x0a   \
    d=\x22m 134.961\
47,95.594487 -79\
.795108,0.0945\x22\x0a\
       style=\x22fi\
ll:#37c8ab;fill-\
rule:evenodd;str\
oke:#aeaeae;stro\
ke-width:16;stro\
ke-linecap:round\
;stroke-linejoin\
:miter;stroke-mi\
terlimit:4;strok\
e-dasharray:none\
;stroke-opacity:\
1\x22 />\x0a  </g>\x0a</s\
vg>\x0a\
"

qt_resource_name = b"\
\x00\x05\
\x00O\xa6S\
\x00I\
\x00c\x00o\x00n\x00s\
\x00\x05\
\x00o\xa6S\
\x00i\
\x00c\x00o\x00n\x00s\
\x00\x14\
\x0b3\xdd\x87\
\x00a\
\x00u\x00t\x00o\x00m\x00a\x00t\x00i\x00c\x00_\x00l\x00a\x00y\x00o\x00u\x00t\x00.\
\x00s\x00v\x00g\
\x00\x0a\
\x01\x0bC\x87\
\x00r\
\x00e\x00s\x00i\x00z\x00e\x00.\x00s\x00v\x00g\
\x00\x0d\
\x07~\x10\x87\
\x00g\
\x00r\x00i\x00d\x00_\x00i\x00c\x00o\x00n\x00.\x00s\x00v\x00g\
\x00\x0f\
\x00 ['\
\x00p\
\x00l\x00u\x00s\x00 \x00(\x00g\x00r\x00a\x00y\x00)\x00.\x00s\x00v\x00g\
\x00\x10\
\x00\x04\x86\xa7\
\x00m\
\x00i\x00n\x00u\x00s\x00 \x00(\x00g\x00r\x00a\x00y\x00)\x00.\x00s\x00v\x00g\
"

qt_resource_struct = b"\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x01\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x00\x00\x02\x00\x00\x00\x01\x00\x00\x00\x02\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\x10\x00\x02\x00\x00\x00\x05\x00\x00\x00\x03\
\x00\x00\x00\x00\x00\x00\x00\x00\
\x00\x00\x00\xac\x00\x00\x00\x00\x00\x01\x00\x00'h\
\x00\x00\x01s\xa14\xaf\xfc\
\x00\x00\x00\x88\x00\x00\x00\x00\x00\x01\x00\x00\x1e8\
\x00\x00\x01s\xa14\xb0\xc8\
\x00\x00\x00N\x00\x00\x00\x00\x00\x01\x00\x00\x07.\
\x00\x00\x01s\xa14\xb0 \
\x00\x00\x00h\x00\x01\x00\x00\x00\x01\x00\x00\x13a\
\x00\x00\x01s\xa14\xb0\xa0\
\x00\x00\x00 \x00\x01\x00\x00\x00\x01\x00\x00\x00\x00\
\x00\x00\x01s\xa14\xb0\x0c\
"

def qInitResources():
    QtCore.qRegisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)

def qCleanupResources():
    QtCore.qUnregisterResourceData(0x03, qt_resource_struct, qt_resource_name, qt_resource_data)

qInitResources()



# GridCal
# Copyright (C) 2022 Santiago Peñate Vera
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
import numpy as np
from PySide2.QtWidgets import *
from PySide2.QtCore import *
from PySide2.QtGui import *

from GridCal.Engine.Devices.transformer3w import Transformer3W
from GridCal.Gui.GridEditorWidget.generic_graphics import ACTIVE, DEACTIVATED, FONT_SCALE, EMERGENCY
from GridCal.Gui.GuiFunctions import ObjectsModel
from GridCal.Engine.Simulations.Topology.topology_driver import reduce_buses
from GridCal.Gui.GridEditorWidget.terminal_item import TerminalItem, HandleItem
from GridCal.Gui.GridEditorWidget.load_graphics import LoadGraphicItem
from GridCal.Gui.GridEditorWidget.generator_graphics import GeneratorGraphicItem
from GridCal.Gui.GridEditorWidget.static_generator_graphics import StaticGeneratorGraphicItem
from GridCal.Gui.GridEditorWidget.battery_graphics import BatteryGraphicItem
from GridCal.Gui.GridEditorWidget.shunt_graphics import ShuntGraphicItem
from GridCal.Gui.GridEditorWidget.messages import *
from GridCal.Engine.Devices.enumerations import DeviceType


class Transformer3WGraphicItem(QGraphicsRectItem):
    """
      Represents a block in the diagram
      Has an x and y and width and height
      width and height can only be adjusted with a tip in the lower right corner.

      - in and output ports
      - parameters
      - description
    """

    def __init__(self, diagramScene, parent=None, index=0, editor=None,
                 elm: Transformer3W = None, pos: QPoint = None, n_windings=3):
        """

        :param diagramScene:
        :param parent:
        :param index:
        :param editor:
        :param elm:
        :param pos:
        :param n_windings:
        """
        QGraphicsRectItem.__init__(self, parent=parent)

        self.min_w = 180.0
        self.min_h = 20.0
        self.offset = 10
        self.h = 70
        self.w = 80
        self.setRect(0.0, 0.0, self.w, self.h)

        self.api_object: Transformer3W = elm
        self.diagramScene = diagramScene  # this is the parent that hosts the pointer to the circuit
        self.editor = editor

        # color
        self.pen_width = 4
        if self.api_object is not None:
            if self.api_object.active:
                self.color = ACTIVE['color']
                self.style = ACTIVE['style']
            else:
                self.color = DEACTIVATED['color']
                self.style = DEACTIVATED['style']
        else:
            self.color = ACTIVE['color']
            self.style = ACTIVE['style']

        self.setPen(QPen(Qt.transparent, self.pen_width, self.style))
        self.setBrush(Qt.transparent)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)
        self.setCursor(QCursor(Qt.PointingHandCursor))

        # index
        self.index = index

        if pos is not None:
            self.setPos(pos)

        # windings
        diameter = self.w * 0.5
        r = diameter / 2
        angle_0 = -90
        d_angle = 360 / n_windings
        angles_deg = [angle_0 + d_angle * i for i in range(n_windings)]
        angles = np.deg2rad(angles_deg)
        x = r * np.cos(angles) + self.w / 4
        y = r * np.sin(angles) + self.w / 4
        xt = diameter * np.cos(angles) + diameter
        yt = diameter * np.sin(angles) + diameter

        self.windings = list()
        self.terminals = list()
        for i in range(n_windings):

            # create objects
            winding = QGraphicsEllipseItem(parent=self)
            winding.setRect(0.0, 0.0, diameter, diameter)
            winding.setPos(x[i], y[i])

            terminal = TerminalItem("t", parent=self, editor=self.editor)
            terminal.setPos(xt[i], yt[i])
            terminal.setRotation(angles_deg[i])

            # set objects style
            winding.setPen(QPen(self.color, self.pen_width, self.style))
            terminal.setPen(QPen(self.color, self.pen_width, self.style))

            self.windings.append(winding)
            self.terminals.append(terminal)

        self.big_marker = None


        # other actions
        self.set_winding_tool_tips()

    def set_winding_tool_tips(self):

        if self.api_object is not None:
            self.windings[0].setToolTip("Winding 1: {0} kV".format(self.api_object.V1))
            self.windings[1].setToolTip("Winding 2: {0} kV".format(self.api_object.V2))
            self.windings[2].setToolTip("Winding 3: {0} kV".format(self.api_object.V3))
        pass

    def set_label(self, val: str):
        """
        Set the label content
        :param val:
        :return:
        """
        self.label.setPlainText(val)

    def mouseMoveEvent(self, event: 'QGraphicsSceneMouseEvent'):
        """
        On mouse move of this object...
        Args:
            event: QGraphicsSceneMouseEvent inherited
        """
        super().mouseMoveEvent(event)

        self.api_object.retrieve_graphic_position()

        # Arrange line positions
        for terminal in self.terminals:
            terminal.process_callbacks(self.pos() + terminal.pos() / 2)

    def add_big_marker(self, color=Qt.red, tool_tip_text=""):
        """
        Add a big marker to the bus
        :param color: Qt Color ot the marker
        :param tool_tip_text: tool tip text to display
        :return:
        """
        if self.big_marker is None:
            self.big_marker = QGraphicsEllipseItem(0, 0, 180, 180, parent=self)
            self.big_marker.setBrush(color)
            self.big_marker.setOpacity(0.5)
            self.big_marker.setToolTip(tool_tip_text)

    def delete_big_marker(self):
        """
        Delete the big marker
        """
        if self.big_marker is not None:
            self.diagramScene.removeItem(self.big_marker)
            self.big_marker = None

    def set_position(self, x, y):
        """
        Set the bus x, y position
        :param x: x in pixels
        :param y: y in pixels
        """
        if np.isnan(x):
            x = 0
        if np.isnan(y):
            y = 0
        self.setPos(QPoint(int(x), int(y)))

    def merge(self, other_bus_graphic):

        self.shunt_children += other_bus_graphic.shunt_children

    def update(self):
        """
        Update the object
        :return:
        """
        pass







import sys
from PySide2.QtWidgets import *
from PySide2 import QtCore, QtGui, QtWidgets


class Wire:

    def __init__(self, name, x, y, gmr, r):
        self.name = name
        self.x = x
        self.y = y
        self.r = r
        self.gmr = gmr


class WiresCollection(QtCore.QAbstractTableModel):

    def __init__(self, parent=None):
        QtCore.QAbstractTableModel.__init__(self, parent)

        self.header = ['Name', 'R (Ohm/km)', 'GMR (m)']

        self.index_prop = {0: 'name', 1: 'r', 2: 'gmr'}

        self.wires = list()

    def add(self, wire: Wire):
        """
        Add wire
        :param wire:
        :return:
        """
        row = len(self.wires)
        self.beginInsertRows(QtCore.QModelIndex(), row, row)
        self.wires.append(wire)
        self.endInsertRows()

    def delete(self, index):
        """
        Delete wire
        :param index:
        :return:
        """
        self.beginRemoveRows(QtCore.QModelIndex(), index, index)
        self.wires.pop(index)
        self.endRemoveRows()

    def rowCount(self, parent=QtCore.QModelIndex()):
        return len(self.wires)

    def columnCount(self, parent=QtCore.QModelIndex()):
        return len(self.header)

    def parent(self, index=None):
        return QtCore.QModelIndex()

    def data(self, index, role=QtCore.Qt.DisplayRole):
        if index.isValid():
            if role == QtCore.Qt.DisplayRole:
                val = getattr(self.wires[index.row()], self.index_prop[index.column()])
                return str(val)
        return None

    def headerData(self, p_int, orientation, role):
        if role == QtCore.Qt.DisplayRole:
            if orientation == QtCore.Qt.Horizontal:
                return self.header[p_int]

    def setData(self, index, value, role=QtCore.Qt.DisplayRole):
        """
        Set data by simple editor (whatever text)
        :param index:
        :param value:
        :param role:
        """
        wire = self.wires[index.row()]
        attr = self.index_prop[index.column()]
        setattr(wire, attr, value)


class TowerBuilderGUI(QtWidgets.QDialog):

    def __init__(self, parent=None):
        """
        Constructor
        Args:
            parent:
        """
        QtWidgets.QDialog.__init__(self, parent)
        self.setWindowTitle('Tower builder')

        # GUI objects
        self.setContextMenuPolicy(QtCore.Qt.NoContextMenu)
        self.layout = QVBoxLayout(self)
        self.wires_tableView = QTableView()
        self.add_wire_pushButton = QPushButton()
        self.add_wire_pushButton.setText('Add')
        self.delete_wire_pushButton = QPushButton()
        self.delete_wire_pushButton.setText('Delete')

        self.layout.addWidget(self.wires_tableView)
        self.layout.addWidget(self.add_wire_pushButton)
        self.layout.addWidget(self.delete_wire_pushButton)

        self.setLayout(self.layout)

        # Model
        self.wire_collection = WiresCollection(self)

        # set models
        self.wires_tableView.setModel(self.wire_collection)

        # button clicks
        self.add_wire_pushButton.clicked.connect(self.add_wire_to_collection)
        self.delete_wire_pushButton.clicked.connect(self.delete_wire_from_collection)

    def msg(self, text, title="Warning"):
        """
        Message box
        :param text: Text to display
        :param title: Name of the window
        """
        msg = QMessageBox()
        msg.setIcon(QMessageBox.Information)
        msg.setText(text)
        # msg.setInformativeText("This is additional information")
        msg.setWindowTitle(title)
        # msg.setDetailedText("The details are as follows:")
        msg.setStandardButtons(QMessageBox.Ok)
        retval = msg.exec_()

    def add_wire_to_collection(self):
        """
        Add new wire to collection
        :return:
        """
        name = 'Wire_' + str(len(self.wire_collection.wires) + 1)
        wire = Wire(name=name, x=0, y=0, gmr=0, r=0.01)
        self.wire_collection.add(wire)

    def delete_wire_from_collection(self):
        """
        Delete wire from the collection
        :return:
        """
        idx = self.wires_tableView.currentIndex()
        sel_idx = idx.row()

        if sel_idx > -1:
            self.wire_collection.delete(sel_idx)
        else:
            self.msg('Select a wire in the wires collection')


if __name__ == "__main__":

    app = QtWidgets.QApplication(sys.argv)
    window = TowerBuilderGUI()
    window.show()
    sys.exit(app.exec_())




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'gui.ui'
##
## Created by: Qt User Interface Compiler version 5.15.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import *
from PySide2.QtGui import *
from PySide2.QtWidgets import *

from .icons_rc import *

class Ui_Dialog(object):
    def setupUi(self, Dialog):
        if not Dialog.objectName():
            Dialog.setObjectName(u"Dialog")
        Dialog.resize(769, 420)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(-1, 0, -1, 9)
        self.frame = QFrame(Dialog)
        self.frame.setObjectName(u"frame")
        self.frame.setMaximumSize(QSize(16777215, 40))
        self.frame.setFrameShape(QFrame.NoFrame)
        self.frame.setFrameShadow(QFrame.Raised)
        self.horizontalLayout_5 = QHBoxLayout(self.frame)
        self.horizontalLayout_5.setSpacing(1)
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.horizontalLayout_5.setContentsMargins(0, 0, 0, 0)
        self.open_button = QPushButton(self.frame)
        self.open_button.setObjectName(u"open_button")
        self.open_button.setMinimumSize(QSize(0, 0))
        icon = QIcon()
        icon.addFile(u":/Icons/icons/import_profiles.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.open_button.setIcon(icon)
        self.open_button.setIconSize(QSize(24, 24))

        self.horizontalLayout_5.addWidget(self.open_button)

        self.refreshButton = QPushButton(self.frame)
        self.refreshButton.setObjectName(u"refreshButton")
        icon1 = QIcon()
        icon1.addFile(u":/Icons/icons/transform.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.refreshButton.setIcon(icon1)
        self.refreshButton.setIconSize(QSize(24, 24))

        self.horizontalLayout_5.addWidget(self.refreshButton)

        self.horizontalSpacer_2 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.horizontalLayout_5.addItem(self.horizontalSpacer_2)


        self.verticalLayout.addWidget(self.frame)

        self.splitter = QSplitter(Dialog)
        self.splitter.setObjectName(u"splitter")
        self.splitter.setOrientation(Qt.Horizontal)
        self.frame_4 = QFrame(self.splitter)
        self.frame_4.setObjectName(u"frame_4")
        self.frame_4.setFrameShape(QFrame.NoFrame)
        self.frame_4.setFrameShadow(QFrame.Raised)
        self.verticalLayout_2 = QVBoxLayout(self.frame_4)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.frame_2 = QFrame(self.frame_4)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setFrameShape(QFrame.NoFrame)
        self.frame_2.setFrameShadow(QFrame.Raised)
        self.formLayout = QFormLayout(self.frame_2)
        self.formLayout.setObjectName(u"formLayout")
        self.formLayout.setContentsMargins(-1, 0, -1, -1)
        self.xComboBox = QComboBox(self.frame_2)
        self.xComboBox.setObjectName(u"xComboBox")

        self.formLayout.setWidget(2, QFormLayout.FieldRole, self.xComboBox)

        self.yComboBox = QComboBox(self.frame_2)
        self.yComboBox.setObjectName(u"yComboBox")

        self.formLayout.setWidget(3, QFormLayout.FieldRole, self.yComboBox)

        self.latitudeComboBox = QComboBox(self.frame_2)
        self.latitudeComboBox.setObjectName(u"latitudeComboBox")

        self.formLayout.setWidget(4, QFormLayout.FieldRole, self.latitudeComboBox)

        self.longitudeComboBox = QComboBox(self.frame_2)
        self.longitudeComboBox.setObjectName(u"longitudeComboBox")

        self.formLayout.setWidget(5, QFormLayout.FieldRole, self.longitudeComboBox)

        self.xCheckBox = QCheckBox(self.frame_2)
        self.xCheckBox.setObjectName(u"xCheckBox")

        self.formLayout.setWidget(2, QFormLayout.LabelRole, self.xCheckBox)

        self.yCheckBox = QCheckBox(self.frame_2)
        self.yCheckBox.setObjectName(u"yCheckBox")

        self.formLayout.setWidget(3, QFormLayout.LabelRole, self.yCheckBox)

        self.latitudeCheckBox = QCheckBox(self.frame_2)
        self.latitudeCheckBox.setObjectName(u"latitudeCheckBox")

        self.formLayout.setWidget(4, QFormLayout.LabelRole, self.latitudeCheckBox)

        self.longitudeCheckBox = QCheckBox(self.frame_2)
        self.longitudeCheckBox.setObjectName(u"longitudeCheckBox")

        self.formLayout.setWidget(5, QFormLayout.LabelRole, self.longitudeCheckBox)

        self.nameRadioButton = QRadioButton(self.frame_2)
        self.nameRadioButton.setObjectName(u"nameRadioButton")
        self.nameRadioButton.setChecked(True)

        self.formLayout.setWidget(0, QFormLayout.LabelRole, self.nameRadioButton)

        self.codeRadioButton = QRadioButton(self.frame_2)
        self.codeRadioButton.setObjectName(u"codeRadioButton")

        self.formLayout.setWidget(1, QFormLayout.LabelRole, self.codeRadioButton)

        self.nameComboBox = QComboBox(self.frame_2)
        self.nameComboBox.setObjectName(u"nameComboBox")

        self.formLayout.setWidget(0, QFormLayout.FieldRole, self.nameComboBox)

        self.codeComboBox = QComboBox(self.frame_2)
        self.codeComboBox.setObjectName(u"codeComboBox")

        self.formLayout.setWidget(1, QFormLayout.FieldRole, self.codeComboBox)


        self.verticalLayout_2.addWidget(self.frame_2)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Minimum, QSizePolicy.Expanding)

        self.verticalLayout_2.addItem(self.verticalSpacer)

        self.splitter.addWidget(self.frame_4)
        self.frame_6 = QFrame(self.splitter)
        self.frame_6.setObjectName(u"frame_6")
        self.frame_6.setFrameShape(QFrame.NoFrame)
        self.frame_6.setFrameShadow(QFrame.Raised)
        self.verticalLayout_3 = QVBoxLayout(self.frame_6)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.assignation_table = QTableView(self.frame_6)
        self.assignation_table.setObjectName(u"assignation_table")

        self.verticalLayout_3.addWidget(self.assignation_table)

        self.frame_9 = QFrame(self.frame_6)
        self.frame_9.setObjectName(u"frame_9")
        self.frame_9.setFrameShape(QFrame.NoFrame)
        self.frame_9.setFrameShadow(QFrame.Raised)
        self.gridLayout = QGridLayout(self.frame_9)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, -1, 0, 0)
        self.horizontalSpacer_3 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer_3, 0, 1, 1, 1)

        self.acceptButton = QPushButton(self.frame_9)
        self.acceptButton.setObjectName(u"acceptButton")
        icon2 = QIcon()
        icon2.addFile(u":/Icons/icons/gear.svg", QSize(), QIcon.Normal, QIcon.Off)
        self.acceptButton.setIcon(icon2)

        self.gridLayout.addWidget(self.acceptButton, 0, 2, 1, 1)


        self.verticalLayout_3.addWidget(self.frame_9)

        self.splitter.addWidget(self.frame_6)

        self.verticalLayout.addWidget(self.splitter)


        self.retranslateUi(Dialog)

        QMetaObject.connectSlotsByName(Dialog)
    # setupUi

    def retranslateUi(self, Dialog):
        Dialog.setWindowTitle(QCoreApplication.translate("Dialog", u"Dialog", None))
#if QT_CONFIG(tooltip)
        self.open_button.setToolTip(QCoreApplication.translate("Dialog", u"Import file", None))
#endif // QT_CONFIG(tooltip)
        self.open_button.setText("")
#if QT_CONFIG(tooltip)
        self.refreshButton.setToolTip(QCoreApplication.translate("Dialog", u"Refresh assignation table", None))
#endif // QT_CONFIG(tooltip)
        self.refreshButton.setText("")
        self.xCheckBox.setText(QCoreApplication.translate("Dialog", u"Bus x position", None))
        self.yCheckBox.setText(QCoreApplication.translate("Dialog", u"Bus y position", None))
        self.latitudeCheckBox.setText(QCoreApplication.translate("Dialog", u"Bus latitude", None))
        self.longitudeCheckBox.setText(QCoreApplication.translate("Dialog", u"Bus longitude", None))
        self.nameRadioButton.setText(QCoreApplication.translate("Dialog", u"Name", None))
        self.codeRadioButton.setText(QCoreApplication.translate("Dialog", u"Code", None))
#if QT_CONFIG(tooltip)
        self.acceptButton.setToolTip(QCoreApplication.translate("Dialog", u"Do it!", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(statustip)
        self.acceptButton.setStatusTip("")
#endif // QT_CONFIG(statustip)
        self.acceptButton.setText(QCoreApplication.translate("Dialog", u"Accept", None))
    # retranslateUi




import subprocess
import sys
import pkg_resources
from GridCal.__version__ import __GridCal_VERSION__


def find_latest_version(name='GridCal'):
    """
    Find the latest version of a package
    :param name: name of the Package
    :return: version string
    """
    latest_version = str(subprocess.run([sys.executable, '-m', 'pip', 'install', '{}==random'.format(name)],
                                        capture_output=True, text=True))
    latest_version = latest_version[latest_version.find('(from versions:') + 15:]
    latest_version = latest_version[:latest_version.find(')')]
    latest_version = latest_version.replace(' ', '').split(',')[-1]
    return latest_version


def check_version():
    """
    Check package version
    :return: version status code, pipy version string

    version status code:
    -2: failure
    -1: this is a newer version
     0: we are ok
    +1: we are behind pipy, we can update
    """

    latest_version = find_latest_version()

    pipy_version = pkg_resources.parse_version(latest_version)
    gc_version = pkg_resources.parse_version(__GridCal_VERSION__)

    if pipy_version is None:
        # could not connect
        return -2, '0.0.0'
    else:
        if hasattr(pipy_version, 'release'):
            if pipy_version.release is None:
                # could not connect
                return -2, '0.0.0'

    if pipy_version == gc_version:
        # same version, we're up to date
        return 0, latest_version

    elif pipy_version > gc_version:
        # we have an older version, we may update
        return 1, latest_version

    elif pipy_version < gc_version:
        # this version is newer than PiPy's
        return -1, latest_version

    else:
        return 0, latest_version


def get_upgrade_command(latest_version=None):
    """
    Get GridCal update command
    :return:
    """
    if latest_version is None:
        latest_version = find_latest_version()
    cmd = [sys.executable, '-m', 'pip', 'install',
           'GridCal=={}'.format(latest_version),
           '--upgrade',
           '--no-dependencies']

    return cmd


if __name__ == '__main__':
    is_latest, curr_ver = check_version()
    print('is the latest', is_latest, curr_ver)
